# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from mapproxy.test.image import is_png, tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "disable_storage.yaml"


class TestDisableStorage(SysTest):

    def test_get_tile_without_caching(self, app, cache_dir):
        for _ in range(2):
            with tmp_image((256, 256), format="png") as img:
                expected_req = (
                    {"path": r"/tile.png"},
                    {"body": img.read(), "headers": {"content-type": "image/png"}},
                )
                with mock_httpd(("localhost", 42423), [expected_req]):
                    resp = app.get("/tms/1.0.0/tiles/0/0/0.png")
                    assert resp.content_type == "image/png"
                    is_png(resp.body)

            assert not cache_dir.check()



# -*- coding: utf-8 -*-
# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Spatial reference systems and transformation of coordinates.
"""
from __future__ import division

import math
import threading
from mapproxy.proj import USE_PROJ4_API
# Old Proj.4 API
from mapproxy.proj import Proj, transform, set_datapath
# New Proj API
from mapproxy.proj import CRS, Transformer
from mapproxy.config import base_config

import logging

from mapproxy.util.bbox import calculate_bbox

log_system = logging.getLogger('mapproxy.system')
log_proj = logging.getLogger('mapproxy.proj')


def get_epsg_num(epsg_code):
    """
    >>> get_epsg_num('ePsG:4326')
    4326
    >>> get_epsg_num(4313)
    4313
    >>> get_epsg_num('31466')
    31466
    >>> get_epsg_num('IGNF:ETRS89UTM28') is None
    True
    """
    if isinstance(epsg_code, str):
        if ':' in epsg_code and epsg_code.upper().startswith('EPSG'):
            epsg_code = int(epsg_code.split(':')[1])
        elif epsg_code.isdigit():
            epsg_code = int(epsg_code)
        else:
            return
    return epsg_code


def get_authority(srs_code):
    """
    >>> get_authority('IAU:1000')
    ('IAU', '1000')
    """
    if isinstance(srs_code, str) and ':' in srs_code:
        auth_name, auth_id = srs_code.rsplit(':', 1)
        return auth_name, auth_id


def _clean_srs_code(code):
    """
    >>> _clean_srs_code(4326)
    'EPSG:4326'
    >>> _clean_srs_code('31466')
    'EPSG:31466'
    >>> _clean_srs_code('crs:84')
    'CRS:84'
    """
    if isinstance(code, str) and ':' in code:
        return code.upper()
    else:
        return 'EPSG:' + str(code)


_proj_initialized = False


def _init_proj():
    global _proj_initialized
    if not _proj_initialized and 'proj_data_dir' in base_config().srs:
        proj_data_dir = base_config().srs['proj_data_dir']
        if proj_data_dir is None:
            _proj_initialized = True
            return
        log_system.info('loading proj data from %s', proj_data_dir)
        set_datapath(proj_data_dir)
        _proj_initialized = True


_thread_local = threading.local()


def SRS(srs_code):
    _init_proj()
    if isinstance(srs_code, _srs_impl):
        return srs_code

    srs_code = _clean_srs_code(srs_code)

    if not hasattr(_thread_local, 'srs_cache'):
        _thread_local.srs_cache = {}

    if srs_code in _thread_local.srs_cache:
        return _thread_local.srs_cache[srs_code]
    else:
        srs = _srs_impl(srs_code)
        _thread_local.srs_cache[srs_code] = srs
        return srs


WEBMERCATOR_EPSG = set(('EPSG:900913', 'EPSG:3857',
                        'EPSG:102100', 'EPSG:102113'))


class _SRS_Proj4_API(object):
    # http://trac.openlayers.org/wiki/SphericalMercator
    proj_init = {
        'EPSG:4326': lambda: Proj('+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs +over'),
        'CRS:84': lambda: Proj('+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs +over'),
    }
    for _epsg in WEBMERCATOR_EPSG:
        proj_init[_epsg] = lambda: Proj(
            '+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 '
            '+lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m '
            '+nadgrids=@null +no_defs +over')

    """
    This class represents a Spatial Reference System.

    Abstracts transformations between different projections.
    Uses the old Proj.4 API, either via pyproj 1 or c-types.
    """

    def __init__(self, srs_code):
        """
        Create a new SRS with the given `srs_code` code.
        """
        self.srs_code = srs_code

        init = self.proj_init.get(srs_code, None)
        if init is not None:
            self.proj = init()
        else:
            epsg_num = get_epsg_num(srs_code)
            if epsg_num is not None:
                self.proj = Proj(init='epsg:%d' % epsg_num)
            else:
                raise ValueError("the old Proj.4 API doesn't support non-EPSG authorities")

    def transform_to(self, other_srs, points):
        """
        :type points: ``(x, y)`` or ``[(x1, y1), (x2, y2), …]``

        >>> srs1 = SRS(4326)
        >>> srs2 = SRS(900913)
        >>> [str(round(x, 5)) for x in srs1.transform_to(srs2, (8.22, 53.15))]
        ['915046.21432', '7010792.20171']
        >>> srs1.transform_to(srs1, (8.25, 53.5))
        (8.25, 53.5)
        >>> [(str(round(x, 5)), str(round(y, 5))) for x, y in
        ...  srs1.transform_to(srs2, [(8.2, 53.1), (8.22, 53.15), (8.3, 53.2)])]
        ... #doctest: +NORMALIZE_WHITESPACE
        [('912819.8245', '7001516.67745'),
         ('915046.21432', '7010792.20171'),
         ('923951.77358', '7020078.53264')]
        """
        if self == other_srs:
            return points
        if isinstance(points[0], (int, float)) and 2 >= len(points) <= 3:
            return transform(self.proj, other_srs.proj, *points)

        x = [p[0] for p in points]
        y = [p[1] for p in points]
        transf_pts = transform(self.proj, other_srs.proj, x, y)
        return zip(transf_pts[0], transf_pts[1])

    def transform_bbox_to(self, other_srs, bbox, with_points=16):
        """

        :param with_points: the number of points to use for the transformation.
            A bbox transformation with only two or four points may cut off some
            parts due to distortions.

        >>> ['%.3f' % x for x in
        ...  SRS(4326).transform_bbox_to(SRS(3857), (-180.0, -90.0, 180.0, 90.0))]
        ['-20037508.343', '-147730762.670', '20037508.343', '147730758.195']
        >>> ['%.5f' % x for x in
        ...  SRS(4326).transform_bbox_to(SRS(3857), (8.2, 53.1, 8.3, 53.2))]
        ['912819.82450', '7001516.67745', '923951.77358', '7020078.53264']
        >>> SRS(4326).transform_bbox_to(SRS(4326), (8.25, 53.0, 8.5, 53.75))
        (8.25, 53.0, 8.5, 53.75)
        """
        if self == other_srs:
            return bbox
        bbox = self.align_bbox(bbox)
        points = generate_envelope_points(bbox, with_points)
        transf_pts = self.transform_to(other_srs, points)
        result = calculate_bbox(transf_pts)

        log_proj.debug('transformed from %r to %r (%s -> %s)',
                       self, other_srs, bbox, result)

        return result

    def align_bbox(self, bbox):
        """
        Align bbox to reasonable values to prevent errors in transformations.
        E.g. transformations from EPSG:4326 with lat=90 or -90 will fail, so
        we subtract a tiny delta.

        At the moment only EPSG:4326 bbox will be modifyed.

        >>> bbox = SRS(4326).align_bbox((-180, -90, 180, 90))
        >>> -90 < bbox[1] < -89.99999998
        True
        >>> 90 > bbox[3] > 89.99999998
        True
        """
        # TODO should not be needed anymore since we transform with +over
        # still a few tests depend on the rounding behavior of this
        if self.srs_code == 'EPSG:4326':
            delta = 0.00000001
            (minx, miny, maxx, maxy) = bbox
            if abs(miny - -90.0) < 1e-6:
                miny = -90.0 + delta
            if abs(maxy - 90.0) < 1e-6:
                maxy = 90.0 - delta
            bbox = minx, miny, maxx, maxy
        return bbox

    @property
    def is_latlong(self):
        """
        >>> SRS(4326).is_latlong
        True
        >>> SRS(31466).is_latlong
        False
        """
        return self.proj.is_latlong()

    def get_geographic_srs(self):
        """ Return the "canonical" geographic CRS corresponding to this CRS.
            Always EPSG:4326 for Proj4 implementation """
        return SRS(4326)

    @property
    def is_axis_order_ne(self):
        """
        Returns `True` if the axis order is North, then East
        (i.e. y/x or lat/lon).

        >>> SRS(4326).is_axis_order_ne
        True
        >>> SRS('CRS:84').is_axis_order_ne
        False
        >>> SRS(31468).is_axis_order_ne
        True
        >>> SRS(31463).is_axis_order_ne
        False
        >>> SRS(25831).is_axis_order_ne
        False
        """
        if self.srs_code in base_config().srs.axis_order_ne:
            return True
        if self.srs_code in base_config().srs.axis_order_en:
            return False
        if self.is_latlong:
            return True
        return False

    @property
    def is_axis_order_en(self):
        """
        Returns `True` if the axis order is East then North
        (i.e. x/y or lon/lat).
        """
        return not self.is_axis_order_ne

    def __eq__(self, other):
        """
        >>> SRS(4326) == SRS("EpsG:4326")
        True
        >>> SRS(4326) == SRS("4326")
        True
        >>> SRS(4326) == SRS(3857)
        False
        """
        if isinstance(other, _SRS_Proj4_API):
            return self.proj.srs == other.proj.srs
        else:
            return NotImplemented

    def __ne__(self, other):
        """
        >>> SRS(3857) != SRS(3857)
        False
        >>> SRS(4326) != SRS(900913)
        True
        """
        equal_result = self.__eq__(other)
        if equal_result is NotImplemented:
            return NotImplemented
        else:
            return not equal_result

    def __str__(self):
        return "SRS %s ('%s')" % (self.srs_code, self.proj.srs)

    def __repr__(self):
        """
        >>> repr(SRS(4326))
        "SRS('EPSG:4326')"
        """
        return "SRS('%s')" % (self.srs_code,)

    def __hash__(self):
        return hash(self.srs_code)


class _SRS(object):
    """
    This class represents a Spatial Reference System.

    Abstracts transformations between different projections.
    Uses the new Proj API via pyproj >=2.
    """

    def __init__(self, srs_code):
        """
        Create a new SRS with the given `srs_code` code.
        """
        self.srs_code = srs_code

        if srs_code in WEBMERCATOR_EPSG:
            epsg_num = 3857
        elif srs_code == 'CRS:84':
            epsg_num = 4326
        else:
            epsg_num = get_epsg_num(srs_code)

        if epsg_num is not None:
            self.proj = CRS.from_epsg(epsg_num)
        else:
            auth_name, auth_id = get_authority(srs_code)
            self.proj = CRS.from_authority(auth_name, auth_id)

        self._transformers = {}

    def _transformer(self, other_srs):
        if other_srs in self._transformers:
            return self._transformers[other_srs]

        t = Transformer.from_crs(self.proj, other_srs.proj, always_xy=True)
        self._transformers[other_srs] = t
        return t

    def transform_to(self, other_srs, points):
        """
        :type points: ``(x, y)`` or ``[(x1, y1), (x2, y2), …]``

        >>> srs1 = SRS(4326)
        >>> srs2 = SRS(900913)
        >>> [str(round(x, 5)) for x in srs1.transform_to(srs2, (8.22, 53.15))]
        ['915046.21432', '7010792.20171']
        >>> srs1.transform_to(srs1, (8.25, 53.5))
        (8.25, 53.5)
        >>> [(str(round(x, 5)), str(round(y, 5))) for x, y in
        ...  srs1.transform_to(srs2, [(8.2, 53.1), (8.22, 53.15), (8.3, 53.2)])]
        ... #doctest: +NORMALIZE_WHITESPACE
        [('912819.8245', '7001516.67745'),
         ('915046.21432', '7010792.20171'),
         ('923951.77358', '7020078.53264')]
        """
        if self == other_srs:
            return points

        transformer = self._transformer(other_srs)
        if isinstance(points[0], (int, float)) and 2 >= len(points) <= 3:
            return transformer.transform(*points)

        x = [p[0] for p in points]
        y = [p[1] for p in points]
        transf_pts = transformer.transform(x, y)
        return zip(transf_pts[0], transf_pts[1])

    def transform_bbox_to(self, other_srs, bbox, with_points=16):
        """

        :param with_points: the number of points to use for the transformation.
            A bbox transformation with only two or four points may cut off some
            parts due to distortions.

        >>> ['%.3f' % x for x in
        ...  SRS(4326).transform_bbox_to(SRS(3857), (-180.0, -90.0, 180.0, 90.0))]
        ['-20037508.343', '-20037508.343', '20037508.343', '20037508.343']
        >>> ['%.5f' % x for x in
        ...  SRS(4326).transform_bbox_to(SRS(3857), (8.2, 53.1, 8.3, 53.2))]
        ['912819.82450', '7001516.67745', '923951.77358', '7020078.53264']
        >>> SRS(4326).transform_bbox_to(SRS(4326), (8.25, 53.0, 8.5, 53.75))
        (8.25, 53.0, 8.5, 53.75)
        """
        if self == other_srs:
            return bbox
        bbox = bbox
        points = generate_envelope_points(bbox, with_points)
        transf_pts = list(self.transform_to(other_srs, points))
        result = calculate_bbox(transf_pts)

        log_proj.debug('transformed from %r to %r (%s -> %s)',
                       self, other_srs, bbox, result)

        # XXX: 3857 is only defined within 85.06 N/S, new Proj returns 'inf' for coords
        # outside of these bounds. Adjust bbox for 4326->3857 transformations to the old
        # behavior, as this is expected in a few places (WMS layer extents and quite a few
        # tests).
        if self.srs_code == 'EPSG:4326' and other_srs.srs_code in ('EPSG:3857', 'EPSG:900913'):
            minx, miny, maxx, maxy = result
            if bbox[0] <= -180.0:
                minx = -20037508.342789244
            if bbox[1] <= -85.06:
                miny = -20037508.342789244
            if bbox[2] >= 180.0:
                maxx = 20037508.342789244
            if bbox[3] >= 85.06:
                maxy = 20037508.342789244
            result = (minx, miny, maxx, maxy)
        return result

    @property
    def is_latlong(self):
        """
        >>> SRS(4326).is_latlong
        True
        >>> SRS(31466).is_latlong
        False
        """
        return self.proj.is_geographic

    def get_geographic_srs(self):
        """ Return the "canonical" geographic CRS corresponding to this CRS.
            EPSG:4326 for Earth CRS, or another one from other celestial bodies """
        auth = self.proj.to_authority()
        if auth is None or not auth[0].startswith('IAU'):
            ret = SRS(4326)
        else:
            return _SRS(':'.join(self.proj.geodetic_crs.to_authority()))
        return ret

    @property
    def is_axis_order_ne(self):
        """
        Returns `True` if the axis order is North, then East
        (i.e. y/x or lat/lon).

        >>> SRS(4326).is_axis_order_ne
        True
        >>> SRS('CRS:84').is_axis_order_ne
        False
        >>> SRS(31468).is_axis_order_ne
        True
        >>> SRS(31463).is_axis_order_ne
        False
        >>> SRS(25831).is_axis_order_ne
        False
        """
        if self.srs_code == 'CRS:84':
            return False
        return self.proj.axis_info[0].direction == 'north'

    @property
    def is_axis_order_en(self):
        """
        Returns `True` if the axis order is East then North
        (i.e. x/y or lon/lat).
        """
        return not self.is_axis_order_ne

    def align_bbox(self, bbox):
        """
        Align bbox to reasonable values to prevent errors in transformations.
        E.g. transformations from EPSG:4326 with lat=90 or -90 will fail, so
        we subtract a tiny delta.

        At the moment only EPSG:4326 bbox will be modifyed.

        >>> bbox = SRS(4326).align_bbox((-180, -90, 180, 90))
        >>> -90 < bbox[1] < -89.99999998
        True
        >>> 90 > bbox[3] > 89.99999998
        True
        """
        # TODO should not be needed anymore since we transform with +over
        # still a few tests depend on the rounding behavior of this
        if self.srs_code == 'EPSG:4326':
            delta = 0.00000001
            (minx, miny, maxx, maxy) = bbox
            if abs(miny - -90.0) < 1e-6:
                miny = -90.0 + delta
            if abs(maxy - 90.0) < 1e-6:
                maxy = 90.0 - delta
            bbox = minx, miny, maxx, maxy
        return bbox

    def __eq__(self, other):
        """
        >>> SRS(4326) == SRS("EpsG:4326")
        True
        >>> SRS(4326) == SRS("4326")
        True
        >>> SRS(4326) == SRS(3857)
        False
        """
        if isinstance(other, _SRS):
            return self.proj.srs == other.proj.srs
        else:
            return NotImplemented

    def __ne__(self, other):
        """
        >>> SRS(3857) != SRS(3857)
        False
        >>> SRS(4326) != SRS(900913)
        True
        """
        equal_result = self.__eq__(other)
        if equal_result is NotImplemented:
            return NotImplemented
        else:
            return not equal_result

    def __str__(self):
        return "SRS %s ('%s')" % (self.srs_code, self.proj.srs)

    def __repr__(self):
        """
        >>> repr(SRS(4326))
        "SRS('EPSG:4326')"
        """
        return "SRS('%s')" % (self.srs_code,)

    def __hash__(self):
        return hash(self.srs_code)

    def to_ogc_url(self):
        """Return a OGC SRS URL like http://www.opengis.net/def/crs/AUTH_NAME/[VERSION]/CODE"""
        auth_name, code = self.srs_code.split(':')
        version = 0
        if auth_name == "OGC":
            version = "1.3"
            if code == "84":
                code = "CRS84"
        if auth_name.startswith("IAU_"):
            version = auth_name[4:]
            auth_name = "IAU"
        return f"http://www.opengis.net/def/crs/{auth_name}/{version}/{code}"

    def semi_major_meters(self):
        """Return the semi major axis in meters of the ellipsoid underlying this SRS"""
        return self.proj.ellipsoid.semi_major_metre


if USE_PROJ4_API:
    _srs_impl = _SRS_Proj4_API
    del _SRS
else:
    _srs_impl = _SRS
    del _SRS_Proj4_API


def generate_envelope_points(bbox, n):
    """
    Generates points that form a linestring around a given bbox.

    @param bbox: bbox to generate linestring for
    @param n: the number of points to generate around the bbox

    >>> generate_envelope_points((10.0, 5.0, 20.0, 15.0), 4)
    [(10.0, 5.0), (20.0, 5.0), (20.0, 15.0), (10.0, 15.0)]
    >>> generate_envelope_points((10.0, 5.0, 20.0, 15.0), 8)
    ... #doctest: +NORMALIZE_WHITESPACE
    [(10.0, 5.0), (15.0, 5.0), (20.0, 5.0), (20.0, 10.0),\
     (20.0, 15.0), (15.0, 15.0), (10.0, 15.0), (10.0, 10.0)]
    """
    (minx, miny, maxx, maxy) = bbox
    if n <= 4:
        n = 0
    else:
        n = int(math.ceil((n - 4) / 4.0))

    width = maxx - minx
    height = maxy - miny

    minx, maxx = min(minx, maxx), max(minx, maxx)
    miny, maxy = min(miny, maxy), max(miny, maxy)

    n += 1
    xstep = width / n
    ystep = height / n
    result = []
    for i in range(n+1):
        result.append((minx + i*xstep, miny))
    for i in range(1, n):
        result.append((maxx, miny + i*ystep))
    for i in range(n, -1, -1):
        result.append((minx + i*xstep, maxy))
    for i in range(n-1, 0, -1):
        result.append((minx, miny + i*ystep))
    return result


def make_lin_transf(src_bbox, dst_bbox):
    """
    Create a transformation function that transforms linear between two
    plane coordinate systems.
    One needs to be cartesian (0, 0 at the lower left, x goes up) and one
    needs to be an image coordinate system (0, 0 at the top left, x goes down).

    :return: function that takes src x/y and returns dest x/y coordinates

    >>> transf = make_lin_transf((7, 50, 8, 51), (0, 0, 500, 400))
    >>> transf((7.5, 50.5))
    (250.0, 200.0)
    >>> transf((7.0, 50.0))
    (0.0, 400.0)
    >>> transf = make_lin_transf((7, 50, 8, 51), (200, 300, 700, 700))
    >>> transf((7.5, 50.5))
    (450.0, 500.0)
    """
    def func(x_y): return (dst_bbox[0] + (x_y[0] - src_bbox[0]) *
                           (dst_bbox[2]-dst_bbox[0]) / (src_bbox[2] - src_bbox[0]),
                           dst_bbox[1] + (src_bbox[3] - x_y[1]) *
                           (dst_bbox[3]-dst_bbox[1]) / (src_bbox[3] - src_bbox[1]))
    return func


class PreferredSrcSRS(object):
    def __init__(self):
        self.target_proj = {}

    def add(self, target, prefered_srs):
        self.target_proj[target] = prefered_srs

    def preferred_src(self, target, available_src):
        if not available_src:
            raise ValueError("no available src SRS")
        if target in available_src:
            return target
        if target in self.target_proj:
            for preferred in self.target_proj[target]:
                if preferred in available_src:
                    return preferred

        for avail in available_src:
            if avail.is_latlong == target.is_latlong:
                return avail
        return available_src[0]


class SupportedSRS(object):
    def __init__(self, supported_srs, preferred_srs=None):
        self.supported_srs = supported_srs
        self.preferred_srs = preferred_srs or PreferredSrcSRS()

    def __iter__(self):
        return iter(self.supported_srs)

    def __contains__(self, srs):
        return srs in self.supported_srs

    def best_srs(self, target):
        return self.preferred_srs.preferred_src(target, self.supported_srs)

    def __eq__(self, other):
        # .prefered_srs is set global, so we only compare .supported_srs
        return self.supported_srs == other.supported_srs


def ogc_crs_url_to_auth_code(url):
    """Convert a OGC CRS URL (http://www.opengis.net/def/crs/AUTH_NAME/[VERSION]/CODE) into 'AUTH_NAME:CODE'"""

    prefix = "http://www.opengis.net/def/crs/"
    if not url.startswith(prefix):
        raise ValueError(f'{url} is not a OGC CRS URL')

    auth_name, version, code = url[len(prefix):].split('/')
    if auth_name == "IAU":
        return auth_name + "_" + version + ":" + code
    return auth_name + ':' + code



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from threading import Lock

from mapproxy.client.http import HTTPClientError
from mapproxy.request.base import BaseRequest
from mapproxy.source import SourceError
from mapproxy.source.wms import WMSLikeSource
from mapproxy.srs import ogc_crs_url_to_auth_code, SRS, SupportedSRS
from mapproxy.util.ogcapi import (
    find_href_in_links,
    normalize_srs_code,
    build_absolute_url,
)
from mapproxy.util.py import reraise_exception

import logging

log = logging.getLogger("mapproxy.source.ogcapimaps")

# For testing
reset_config_cache = False


class OGCAPIMapsSource(WMSLikeSource):
    def __init__(
        self,
        landingpage_url,
        collection,
        http_client,
        coverage=None,
        image_opts=None,
        error_handler=None,
        res_range=None,
        supported_srs=None,
        transparent=None,
        transparent_color=None,
        transparent_color_tolerance=None,
        bgcolor=None,
    ):
        WMSLikeSource.__init__(
            self,
            image_opts=image_opts,
            coverage=coverage,
            res_range=res_range,
            transparent_color=transparent_color,
            transparent_color_tolerance=transparent_color_tolerance,
            error_handler=error_handler,
        )

        self.landingpage_url = landingpage_url.rstrip("/")
        self.collection = collection
        self.http_client = http_client
        if transparent is not None:
            self.image_opts.transparent = transparent
        self.bgcolor = bgcolor
        self.supported_srs = supported_srs
        self.supported_srs_from_config = (
            self.supported_srs is not None and len(self.supported_srs.supported_srs) > 0
        )

        self.lock = Lock()
        # Below variables are protected under the lock
        self._reset_config_caches()

    def _reset_config_caches(self):
        self.has_got_maps_list = False
        self.map_format_to_href = {}
        if not self.supported_srs_from_config:
            self.supported_srs = set()
        self.map_crs_code_to_url = {}

    def _build_url(self, href):
        return build_absolute_url(self.landingpage_url, href)

    def _get_maps_list(self):
        with self.lock:
            # Used by tests to avoid caching of "metadata"
            global reset_config_cache
            if reset_config_cache:
                reset_config_cache = False
                self._reset_config_caches()

            if self.has_got_maps_list:
                return
            self.has_got_maps_list = True

            headers = {"Accept": "application/json"}

            url = self.landingpage_url
            if self.collection:
                url += "/collections/" + self.collection
            try:
                resp = self.http_client.open(url, headers=headers)
            except HTTPClientError as e:
                log.warning(f"Cannot retrieve {url}: %s", e)
                reraise_exception(SourceError(e.args[0]), sys.exc_info())
            try:
                j = json.loads(resp.read().decode("utf-8"))
            except Exception as e:
                log.warning(f"Cannot parse response to {url} as JSON: %s", e)
                reraise_exception(SourceError(e.args[0]), sys.exc_info())

            if "links" not in j:
                ex = SourceError(f"Could not retrieve 'links' in {url} response")
                log.error(ex)
                raise ex

            for mimetype in ("image/png", "image/jpeg"):
                href = find_href_in_links(
                    j["links"],
                    "http://www.opengis.net/def/rel/ogc/1.0/map",
                    mimetype,
                )
                if href:
                    len_image_prefix = len("image/")
                    self.map_format_to_href[
                        mimetype[len_image_prefix:]
                    ] = self._build_url(href)

            if len(self.map_format_to_href) == 0:
                ex = SourceError(f"Could not retrieve a map link in {url} response")
                log.error(ex)
                raise ex

            supported_srs = set()
            if "crs" in j:
                for crs_url in j["crs"]:
                    crs_code = normalize_srs_code(ogc_crs_url_to_auth_code(crs_url))
                    self.map_crs_code_to_url[crs_code] = crs_url
                    supported_srs.add(crs_code)
            if "storageCrs" in j:
                crs_url = j["storageCrs"]
                crs_code = normalize_srs_code(ogc_crs_url_to_auth_code(crs_url))
                self.map_crs_code_to_url[crs_code] = crs_url
                supported_srs.add(crs_code)

            if len(supported_srs) == 0:
                self.map_crs_code_to_url[
                    "EPSG:4326"
                ] = "http://www.opengis.net/def/crs/OGC/1.3/CRS84"
                supported_srs.add("EPSG:4326")

            if self.supported_srs_from_config:
                for srs in self.supported_srs.supported_srs:
                    self.map_crs_code_to_url[srs.srs_code] = srs.to_ogc_url()
            else:
                self.supported_srs = SupportedSRS([SRS(crs) for crs in supported_srs])

    def get_map(self, query):
        self._get_maps_list()
        return WMSLikeSource.get_map(self, query)

    def _retrieve(self, query, format):
        url = self.map_format_to_href[format]
        req = BaseRequest(url=url)
        if query.srs.srs_code != "EPSG:4326" and query.srs.is_axis_order_ne:
            req.params["bbox"] = "%.17g,%.17g,%.17g,%.17g" % (
                query.bbox[1],
                query.bbox[0],
                query.bbox[3],
                query.bbox[2],
            )
        else:
            req.params["bbox"] = "%.17g,%.17g,%.17g,%.17g" % (
                query.bbox[0],
                query.bbox[1],
                query.bbox[2],
                query.bbox[3],
            )
        if query.srs.srs_code != "EPSG:4326":
            req.params["bbox-crs"] = self.map_crs_code_to_url[query.srs.srs_code]
        if (
            query.srs.srs_code != "EPSG:4326"
            or len(self.supported_srs.supported_srs) != 1
        ):
            req.params["crs"] = self.map_crs_code_to_url[query.srs.srs_code]
        req.params["width"] = query.size[0]
        req.params["height"] = query.size[1]
        if self.image_opts.transparent:
            req.params["transparent"] = "true"
        if self.bgcolor:
            if len(self.bgcolor) == 4:
                r, g, b, a = self.bgcolor
                req.params["bgcolor"] = "0x%02X%02X%02X%02X" % (a, r, g, b)
            else:
                r, g, b = self.bgcolor
                req.params["bgcolor"] = "0x%02X%02X%02X" % (r, g, b)
        resp = self.http_client.open(req.complete_url)
        return resp



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Python related helper functions.
"""
from functools import wraps


def reraise_exception(new_exc, exc_info):
    """
    Reraise exception (`new_exc`) with the given `exc_info`.
    """
    _exc_class, _exc, tb = exc_info
    raise new_exc.with_traceback(tb)


def reraise(exc_info):
    """
    Reraise exception from exc_info`.
    """
    exc_class, exc, tb = exc_info
    raise exc.with_traceback(tb)


class cached_property(object):
    """A decorator that converts a function into a lazy property. The
    function wrapped is called the first time to retrieve the result
    and than that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42
    """

    def __init__(self, func, name=None, doc=None):
        self.func = func
        self.__name__ = name or func.__name__
        self.__doc__ = doc or func.__doc__

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = self.func(obj)
        setattr(obj, self.__name__, value)
        return value


def memoize(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, '__memoize_cache'):
            self.__memoize_cache = {}
        cache = self.__memoize_cache.setdefault(func, {})
        key = args + tuple(kwargs.items())
        if key not in cache:
            cache[key] = func(self, *args, **kwargs)
        return cache[key]
    return wrapper



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import shutil
import tempfile
import os
import re
import sys
from glob import glob as globfunc
from contextlib import contextmanager
from lxml import etree

from mapproxy.test import mocker


class Mocker(object):
    """
    This is a base class for unit-tests that use ``mocker``. This class follows
    the xUnit naming conventions for setup and teardown methods.

    `setup` will initialize a `mocker.Mocker`. The `teardown` method
    will run ``mocker.verify()``.
    """

    def setup_method(self):
        self.mocker = mocker.Mocker()

    def expect_and_return(self, mock_call, return_val):
        """
        Register a return value for the mock call.
        :param return_val: The value mock_call should return.
        """
        self.mocker.result(return_val)

    def expect(self, mock_call):
        return mocker.expect(mock_call)

    def replay(self):
        """
        Finish mock-record phase.
        """
        self.mocker.replay()

    def mock(self, base_cls=None):
        """
        Return a new mock object.
        :param base_cls: check method signatures of the mock-calls with this
            base_cls signature (optional)
        """
        if base_cls:
            return self.mocker.mock(base_cls)
        return self.mocker.mock()

    def teardown_method(self):
        self.mocker.verify()


class TempFiles(object):
    """
    This class is a context manager for temporary files.

    >>> with TempFiles(n=2, suffix='.png') as tmp:
    ...     for f in tmp:
    ...         assert os.path.exists(f)
    >>> for f in tmp:
    ...     assert not os.path.exists(f)
    """

    def __init__(self, n=1, suffix='', no_create=False):
        self.n = n
        self.suffix = suffix
        self.no_create = no_create
        self.tmp_files = []

    def __enter__(self):
        for _ in range(self.n):
            fd, tmp_file = tempfile.mkstemp(suffix=self.suffix)
            os.close(fd)
            self.tmp_files.append(tmp_file)
            if self.no_create:
                os.remove(tmp_file)
        return self.tmp_files

    def __exit__(self, exc_type, exc_val, exc_tb):
        for tmp_file in self.tmp_files:
            if os.path.exists(tmp_file):
                os.remove(tmp_file)
        self.tmp_files = []


class TempFile(TempFiles):
    def __init__(self, suffix='', no_create=False):
        TempFiles.__init__(self, suffix=suffix, no_create=no_create)

    def __enter__(self):
        return TempFiles.__enter__(self)[0]


class TempDir:
    def __enter__(self):
        self.tmp_dir = tempfile.mkdtemp()
        return self.tmp_dir

    def __exit__(self, exc_type, exc_val, exc_tb):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir, ignore_errors=True)


class ChangeWorkingDir:
    def __init__(self, new_dir):
        self.new_dir = new_dir

    def __enter__(self):
        self.old_dir = os.getcwd()
        os.chdir(self.new_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.old_dir)


class LogMock(object):
    log_methods = ('info', 'debug', 'warn', 'error', 'fail')

    def __init__(self, module, log_name='log'):
        self.module = module
        self.orig_logger = None
        self.logged_msgs = []

    def __enter__(self):
        self.orig_logger = self.module.log
        self.module.log = self
        return self

    def __getattr__(self, name):
        if name in self.log_methods:
            def _log(msg):
                self.logged_msgs.append((name, msg))
            return _log
        raise AttributeError("'%s' object has no attribute '%s'" %
                             (self.__class__.__name__, name))

    def assert_log(self, type, msg):
        log_type, log_msg = self.logged_msgs.pop(0)
        assert log_type == type, 'expected %s log message, but was %s' % (type, log_type)
        assert msg in log_msg.lower(), "expected string '%s' in log message '%s'" % \
            (msg, log_msg)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.module.log = self.orig_logger


def assert_re(value, regex):
    """
    >>> assert_re('hello', 'l+')
    >>> assert_re('hello', 'l{3}')
    Traceback (most recent call last):
        ...
    AssertionError: hello ~= l{3}
    """
    match = re.search(regex, value)
    assert match is not None, '%s ~= %s' % (value, regex)


def assert_files_in_dir(dir, expected, glob=None):
    """
    assert that (only) ``expected`` files are in ``dir``.
    ``filter`` can be a globbing patter, other files are ignored if it is set.
    """
    if glob is not None:
        files = globfunc(os.path.join(dir, glob))
        files = [os.path.basename(f) for f in files]
    else:
        files = os.listdir(dir)
    files.sort()
    sorted_expected = sorted(expected)
    assert sorted_expected == files, f'{", ".join(sorted_expected)} ~= {", ".join(files)}'


def validate_with_dtd(doc, dtd_name, dtd_basedir=None):
    if dtd_basedir is None:
        dtd_basedir = os.path.join(os.path.dirname(__file__), 'schemas')

    dtd_filename = os.path.join(dtd_basedir, dtd_name)
    with open(dtd_filename, 'rb') as schema:
        dtd = etree.DTD(schema)
        if isinstance(doc, (str, bytes)):
            xml = etree.XML(doc)
        else:
            xml = doc
        is_valid = dtd.validate(xml)
        print(dtd.error_log.filter_from_errors())
        return is_valid


def validate_with_xsd(doc, xsd_name, xsd_basedir=None):
    if xsd_basedir is None:
        xsd_basedir = os.path.join(os.path.dirname(__file__), 'schemas')

    xsd_filename = os.path.join(xsd_basedir, xsd_name)

    with open(xsd_filename, 'rb') as schema:
        xsd = etree.parse(schema)
        xml_schema = etree.XMLSchema(xsd)
        if isinstance(doc, (str, bytes)):
            xml = etree.XML(doc)
        else:
            xml = doc
        is_valid = xml_schema.validate(xml)
        print(xml_schema.error_log.filter_from_errors())
        return is_valid


class XPathValidator(object):
    def __init__(self, doc):
        self.xml = etree.XML(doc)

    def assert_xpath(self, xpath, expected=None):
        assert len(self.xml.xpath(xpath)) > 0, xpath + ' does not match anything'
        if expected is not None:
            if callable(expected):
                assert expected(self.xml.xpath(xpath)[0])
            else:
                assert self.xml.xpath(xpath)[0] == expected

    def xpath(self, xpath):
        return self.xml.xpath(xpath)


def strip_whitespace(data):
    """
    >>> strip_whitespace(' <foo> bar\\n zing\\t1')
    '<foo>barzing1'
    """
    if isinstance(data, bytes):
        return re.sub(br'\s+', b'', data)
    else:
        return re.sub(r'\s+', '', data)


@contextmanager
def capture(bytes=False):
    if bytes:
        from io import BytesIO as StringIO
    else:
        from io import StringIO

    backup_stdout = sys.stdout
    backup_stderr = sys.stderr

    try:
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        yield sys.stdout, sys.stderr
    except Exception as ex:
        backup_stdout.write(str(ex))
        if bytes:
            backup_stdout.write(sys.stdout.getvalue().decode('utf-8'))
            backup_stderr.write(sys.stderr.getvalue().decode('utf-8'))
        else:
            backup_stdout.write(sys.stdout.getvalue())
            backup_stderr.write(sys.stderr.getvalue())
        raise
    finally:
        sys.stdout = backup_stdout
        sys.stderr = backup_stderr


def assert_permissions(file_path, permissions):
    actual_permissions = oct(os.stat(file_path).st_mode & 0o777)
    desired_permissions = oct(int(permissions, base=8))
    assert actual_permissions == desired_permissions, f'{actual_permissions} ~= {desired_permissions}'



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.image.opts import ImageOptions
from mapproxy.image import BlankImageSource


class HTTPSourceErrorHandler(object):
    def __init__(self):
        self.response_error_codes = {}

    def add_handler(self, http_code, color, cacheable=False, authorize_stale=False):
        self.response_error_codes[http_code] = (color, cacheable, authorize_stale)

    def handle(self, status_code, query):
        color = cacheable = None
        if status_code in self.response_error_codes:
            color, cacheable, authorize_stale = self.response_error_codes[status_code]
        elif 'other' in self.response_error_codes:
            color, cacheable, authorize_stale = self.response_error_codes['other']
        else:
            return None

        transparent = len(color) == 4
        image_opts = ImageOptions(bgcolor=color, transparent=transparent)
        img_source = BlankImageSource(query.size, image_opts, cacheable=cacheable)
        img_source.authorize_stale = authorize_stale
        return img_source



class TransformationError(Exception):
    pass


def calculate_bbox(points):
    """
    Calculates the bbox of a list of points.

    >>> calculate_bbox([(-5, 20), (3, 8), (99, 0)])
    (-5, 0, 99, 20)

    @param points: list of points [(x0, y0), (x1, y2), ...]
    @returns: bbox of the input points.
    """
    points = list(points)
    # points can be INF for invalid transformations, filter out
    try:
        minx = min(p[0] for p in points if p[0] != float('inf'))
        miny = min(p[1] for p in points if p[1] != float('inf'))
        maxx = max(p[0] for p in points if p[0] != float('inf'))
        maxy = max(p[1] for p in points if p[1] != float('inf'))
        return (minx, miny, maxx, maxy)
    except ValueError:  # min/max are called with empty list when everything is inf
        raise TransformationError()


def merge_bbox(bbox1, bbox2):
    """
    Merge two bboxes.

    >>> merge_bbox((-10, 20, 0, 30), (30, -20, 90, 10))
    (-10, -20, 90, 30)

    """
    minx = min(bbox1[0], bbox2[0])
    miny = min(bbox1[1], bbox2[1])
    maxx = max(bbox1[2], bbox2[2])
    maxy = max(bbox1[3], bbox2[3])
    return (minx, miny, maxx, maxy)


def bbox_equals(src_bbox, dst_bbox, x_delta=None, y_delta=None):
    """
    Compares two bbox and checks if they are equal, or nearly equal.

    :param x_delta: how precise the comparison should be.
                    should be reasonable small, like a tenth of a pixel.
                    defaults to 1/1.000.000th of the width.
    :type x_delta: bbox units

    >>> src_bbox = (939258.20356824622, 6887893.4928338043,
    ...             1095801.2374962866, 7044436.5267618448)
    >>> dst_bbox = (939258.20260000182, 6887893.4908000007,
    ...             1095801.2365000017, 7044436.5247000009)
    >>> bbox_equals(src_bbox, dst_bbox, 61.1, 61.1)
    True
    >>> bbox_equals(src_bbox, dst_bbox, 0.0001)
    False
    """
    if x_delta is None:
        x_delta = abs(src_bbox[0] - src_bbox[2]) / 1000000.0
    if y_delta is None:
        y_delta = x_delta
    return (abs(src_bbox[0] - dst_bbox[0]) < x_delta and
            abs(src_bbox[1] - dst_bbox[1]) < x_delta and
            abs(src_bbox[2] - dst_bbox[2]) < y_delta and
            abs(src_bbox[3] - dst_bbox[3]) < y_delta)


def bbox_tuple(bbox):
    """
    >>> bbox_tuple('20,-30,40,-10')
    (20.0, -30.0, 40.0, -10.0)
    >>> bbox_tuple([20,-30,40,-10])
    (20.0, -30.0, 40.0, -10.0)

    """
    if isinstance(bbox, str):
        bbox = bbox.split(',')
    bbox = tuple(map(float, bbox))
    return bbox


def bbox_width(bbox):
    return bbox[2] - bbox[0]


def bbox_height(bbox):
    return bbox[3] - bbox[1]


def bbox_size(bbox):
    return bbox_width(bbox), bbox_height(bbox)


def bbox_intersects(one, two):
    a_x0, a_y0, a_x1, a_y1 = one
    b_x0, b_y0, b_x1, b_y1 = two

    if (
            a_x0 < b_x1 and
            a_x1 > b_x0 and
            a_y0 < b_y1 and
            a_y1 > b_y0
    ):
        return True

    return False


def bbox_contains(one, two):
    """
    Returns ``True`` if `one` contains `two`.

    >>> bbox_contains([0, 0, 10, 10], [2, 2, 4, 4])
    True
    >>> bbox_contains([0, 0, 10, 10], [0, 0, 11, 10])
    False

    Allow tiny rounding errors:

    >>> bbox_contains([0, 0, 10, 10], [0.000001, 0.0000001, 10.000001, 10.000001])
    False
    >>> bbox_contains([0, 0, 10, 10], [0.0000000000001, 0.0000000000001, 10.0000000000001, 10.0000000000001])
    True
    """
    a_x0, a_y0, a_x1, a_y1 = one
    b_x0, b_y0, b_x1, b_y1 = two

    x_delta = abs(a_x1 - a_x0) / 10e12
    y_delta = abs(a_y1 - a_y0) / 10e12

    if (
            a_x0 <= b_x0 + x_delta and
            a_x1 >= b_x1 - x_delta and
            a_y0 <= b_y0 + y_delta and
            a_y1 >= b_y1 - y_delta
    ):
        return True

    return False


def bbox_from_center(center_x, center_y, width, height):
    """Compute a bounding box from its center, and its width and height
    """
    return (
        center_x - width / 2,
        center_y - height / 2,
        center_x + width / 2,
        center_y + height / 2,
    )



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import time
import sys

import pytest

from mapproxy.seed.cachelock import CacheLocker, CacheLockedError


@pytest.mark.skipif(sys.platform == "win32", reason="test not supported for Windows")
@pytest.mark.skipif(sys.platform == "darwin", reason="test not supported for MacOS")
class TestCacheLock(object):

    @pytest.fixture
    def lock_file(self, tmpdir):
        return (tmpdir / "lock").strpath

    def test_free_lock(self, lock_file):
        locker = CacheLocker(lock_file)
        with locker.lock("foo"):
            assert True

    def test_locked_by_process_no_block(self, lock_file):
        proc_is_locked = multiprocessing.Event()

        def lock():
            locker = CacheLocker(lock_file)
            with locker.lock("foo"):
                proc_is_locked.set()
                time.sleep(10)

        p = multiprocessing.Process(target=lock)
        p.start()
        # wait for process to start
        proc_is_locked.wait()

        locker = CacheLocker(lock_file)

        # test unlocked bar
        with locker.lock("bar", no_block=True):
            assert True

        # test locked foo
        try:
            with locker.lock("foo", no_block=True):
                assert False
        except CacheLockedError:
            pass
        finally:
            p.terminate()
            p.join()

    def test_locked_by_process_waiting(self, lock_file):
        proc_is_locked = multiprocessing.Event()

        def lock():
            locker = CacheLocker(lock_file)
            with locker.lock("foo"):
                proc_is_locked.set()
                time.sleep(.1)

        p = multiprocessing.Process(target=lock)
        start_time = time.time()
        p.start()
        # wait for process to start
        proc_is_locked.wait()

        locker = CacheLocker(lock_file, polltime=0.02)
        try:
            with locker.lock("foo", no_block=False):
                diff = time.time() - start_time
                assert diff > 0.1
        finally:
            p.terminate()
            p.join()



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import sys
import optparse
from io import BytesIO
from urllib.parse import urlparse

from xml import etree

from mapproxy.client.http import open_url, HTTPClientError
from mapproxy.request.base import BaseRequest, url_decode
from mapproxy.util.ext import wmsparse


class PrettyPrinter(object):
    def __init__(self, indent=4, version='1.1.1'):
        self.indent = indent
        self.print_order = ['name', 'title', 'url', 'srs', 'llbbox', 'bbox']
        self.marker = '- '
        self.version = version

    def print_line(self, indent, key, value=None, mark_first=False):
        marker = ''
        if value is None:
            value = ''
        if mark_first:
            indent = indent - len(self.marker)
            marker = self.marker
        print(("%s%s%s: %s" % (' '*indent, marker, key, value)))

    def _format_output(self, key, value, indent, mark_first=False):
        if key == 'bbox':
            self.print_line(indent, key)
            for srs_code, bbox in value.items():
                self.print_line(indent+self.indent, srs_code, value=bbox, mark_first=mark_first)
        else:
            if isinstance(value, set):
                value = list(value)
            self.print_line(indent, key, value=value, mark_first=mark_first)

    def print_layers(self, capabilities, indent=None, root=False):
        if root:
            print("# Note: This is not a valid MapProxy configuration!")
            print('Capabilities Document Version %s' % (self.version,))
            print('Root-Layer:')
            layer_list = capabilities.layers()['layers']
        else:
            layer_list = capabilities['layers']

        indent = indent or self.indent
        for layer in layer_list:
            marked_first = False
            # print ordered items first
            for item in self.print_order:
                if layer.get(item, False):
                    if not marked_first:
                        marked_first = True
                        self._format_output(item, layer[item], indent, mark_first=marked_first)
                    else:
                        self._format_output(item, layer[item], indent)
            # print remaining items except sublayers
            for key, value in layer.items():
                if key in self.print_order or key == 'layers':
                    continue
                self._format_output(key, value, indent)
            # print the sublayers now
            if layer.get('layers', False):
                self.print_line(indent, 'layers')
                self.print_layers(layer, indent=indent+self.indent)


def log_error(msg, *args):
    print(msg % args, file=sys.stderr)


def wms_capapilities_url(url, version):
    parsed_url = urlparse(url)
    base_req = BaseRequest(
        url=url.split('?', 1)[0],
        param=url_decode(parsed_url.query),
    )

    base_req.params['service'] = 'WMS'
    base_req.params['version'] = version
    base_req.params['request'] = 'GetCapabilities'
    return base_req.complete_url


def parse_capabilities(fileobj, version='1.1.1'):
    try:
        return wmsparse.parse_capabilities(fileobj)
    except ValueError as ex:
        log_error('%s\n%s\n%s\n%s\nNot a capabilities document: %s',
                  'Recieved document:', '-'*80, fileobj.getvalue(), '-'*80, ex.args[0])
        sys.exit(1)
    except etree.ElementTree.ParseError as ex:
        log_error('%s\n%s\n%s\n%s\nCould not parse the document: %s',
                  'Recieved document:', '-'*80, fileobj.getvalue(), '-'*80, ex.args[0])
        sys.exit(1)


def parse_capabilities_url(url, version='1.1.1'):
    try:
        capabilities_url = wms_capapilities_url(url, version)
        capabilities_response = open_url(capabilities_url)
    except HTTPClientError as ex:
        log_error('ERROR: %s', ex.args[0])
        sys.exit(1)

    # after parsing capabilities_response will be empty, therefore cache it
    capabilities = BytesIO(capabilities_response.read())
    return parse_capabilities(capabilities, version=version)


def wms_capabilities_command(args=None):
    parser = optparse.OptionParser("%prog wms-capabilities [options] URL",
                                   description="Read and parse WMS 1.1.1 capabilities and print out"
                                   " information about each layer. It does _not_ return a valid"
                                   " MapProxy configuration.")
    parser.add_option("--host", dest="capabilities_url",
                      help="WMS Capabilites URL")
    parser.add_option(
        "--version", dest="version", choices=['1.1.1', '1.3.0'], default='1.1.1',
        help="Request GetCapabilities-document in version 1.1.1 or 1.3.0", metavar="<1.1.1 or 1.3.0>")

    if args:
        args = args[1:]  # remove script name

    (options, args) = parser.parse_args(args)
    if not options.capabilities_url:
        if len(args) != 1:
            parser.print_help()
            sys.exit(2)
        else:
            options.capabilities_url = args[0]

    try:
        service = parse_capabilities_url(options.capabilities_url, version=options.version)

        printer = PrettyPrinter(indent=4, version=options.version)
        printer.print_layers(service, root=True)

    except KeyError as ex:
        log_error('XML-Element has no such attribute (%s).' % (ex.args[0],))
        sys.exit(1)



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import errno
import os
import time
from contextlib import contextmanager

from mapproxy.util.sqlite3 import sqlite3


class CacheLockedError(Exception):
    pass


class CacheLocker(object):
    def __init__(self, lockfile, polltime=0.1):
        self.lockfile = lockfile
        self.polltime = polltime
        self._initialize_lockfile()

    def _initialize_lockfile(self):
        db = sqlite3.connect(self.lockfile)
        db.execute("""
            CREATE TABLE IF NOT EXISTS cache_locks (
                cache_name TEXT NOT NULL,
                created REAL NOT NULL,
                pid INTEGER NUT NULL
            );
        """)
        db.commit()
        db.close()

    @contextmanager
    def _exclusive_db_cursor(self):
        db = sqlite3.connect(self.lockfile, isolation_level="EXCLUSIVE")
        db.row_factory = sqlite3.Row
        cur = db.cursor()

        try:
            yield cur
        finally:
            db.commit()
            db.close()

    @contextmanager
    def lock(self, cache_name, no_block=False):

        pid = os.getpid()

        while True:
            with self._exclusive_db_cursor() as cur:
                self._add_lock(cur, cache_name, pid)
                if self._poll(cur, cache_name, pid):
                    break
                elif no_block:
                    raise CacheLockedError()
            time.sleep(self.polltime)

        try:
            yield
        finally:
            with self._exclusive_db_cursor() as cur:
                self._remove_lock(cur, cache_name, pid)

    def _poll(self, cur, cache_name, pid):
        active_locks = False
        cur.execute("SELECT * from cache_locks where cache_name = ? ORDER BY created", (cache_name, ))

        for lock in cur:
            if not active_locks and lock['cache_name'] == cache_name and lock['pid'] == pid:
                # we are waiting and it is out turn
                return True

            if not is_running(lock['pid']):
                self._remove_lock(cur, lock['cache_name'], lock['pid'])
            else:
                active_locks = True

        return not active_locks

    def _add_lock(self, cur, cache_name, pid):
        cur.execute("SELECT count(*) from cache_locks WHERE cache_name = ? AND pid = ?", (cache_name, pid))
        if cur.fetchone()[0] == 0:
            cur.execute("INSERT INTO cache_locks (cache_name, pid, created) VALUES (?, ?, ?)",
                        (cache_name, pid, time.time()))

    def _remove_lock(self, cur, cache_name, pid):
        cur.execute("DELETE FROM cache_locks WHERE cache_name = ?  AND pid = ?", (cache_name, pid))


class DummyCacheLocker(object):
    @contextmanager
    def lock(self, cache_name, no_block=False):
        yield


def is_running(pid):
    try:
        os.kill(pid, 0)
    except OSError as err:
        if err.errno == errno.ESRCH:
            return False
        elif err.errno == errno.EPERM:
            return True
        else:
            raise err
    else:
        return True


if __name__ == '__main__':
    locker = CacheLocker('/tmp/cachelock_test')
    with locker.lock('foo'):
        pass



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial as fp
from urllib.parse import urlsplit, urlunsplit

from mapproxy.request.base import RequestParams, BaseRequest
from mapproxy.srs import make_lin_transf


class ArcGISExportRequestParams(RequestParams):
    """
    Supported params f, bbox(required), size, dpi, imageSR, bboxSR, format, layerDefs,
    layers, transparent, time, layerTimeOptions.

    @param layers: Determines which layers appear on the exported map. There are
                   four ways to specify layers: show, hide, include, exclude.
                   (ex show:1,2)
    """

    def _get_format(self):
        """
        The requested format as string (w/o any 'image/', 'text/', etc prefixes)
        """
        return self["format"]

    def _set_format(self, format):
        self["format"] = format.rsplit("/")[-1]
    format = property(_get_format, _set_format)
    del _get_format
    del _set_format

    def _get_bbox(self):
        """
        ``bbox`` as a tuple (minx, miny, maxx, maxy).
        """
        if 'bbox' not in self.params or self.params['bbox'] is None:
            return None
        points = [float(val) for val in self.params['bbox'].split(',')]
        return tuple(points[:4])

    def _set_bbox(self, value):
        if value is not None and not isinstance(value, str):
            value = ','.join(str(x) for x in value)
        self['bbox'] = value
    bbox = property(_get_bbox, _set_bbox)
    del _get_bbox
    del _set_bbox

    def _get_size(self):
        """
        Size of the request in pixel as a tuple (width, height),
        or None if one is missing.
        """
        if 'size' not in self.params or self.params['size'] is None:
            return None
        dim = [float(val) for val in self.params['size'].split(',')]
        return tuple(dim[:2])

    def _set_size(self, value):
        if value is not None and not isinstance(value, str):
            value = ','.join(str(x) for x in value)
        self['size'] = value
    size = property(_get_size, _set_size)
    del _get_size
    del _set_size

    def _get_srs(self, key):
        return self.params.get(key, None)

    def _set_srs(self, srs, key):
        if hasattr(srs, 'srs_code'):
            code = srs.srs_code
        else:
            code = srs
        self.params[key] = code.rsplit(":", 1)[-1]

    bboxSR = property(fp(_get_srs, key="bboxSR"), fp(_set_srs, key="bboxSR"))
    imageSR = property(fp(_get_srs, key="imageSR"), fp(_set_srs, key="imageSR"))
    del _get_srs
    del _set_srs


class ArcGISIdentifyRequestParams(ArcGISExportRequestParams):
    def _get_format(self):
        """
        The requested format as string (w/o any 'image/', 'text/', etc prefixes)
        """
        return self["format"]

    def _set_format(self, format):
        self["format"] = format.rsplit("/")[-1]
    format = property(_get_format, _set_format)
    del _get_format
    del _set_format

    def _get_bbox(self):
        """
        ``bbox`` as a tuple (minx, miny, maxx, maxy).
        """
        if 'mapExtent' not in self.params or self.params['mapExtent'] is None:
            return None
        points = [float(val) for val in self.params['mapExtent'].split(',')]
        return tuple(points[:4])

    def _set_bbox(self, value):
        if value is not None and not isinstance(value, str):
            value = ','.join(str(x) for x in value)
        self['mapExtent'] = value
    bbox = property(_get_bbox, _set_bbox)
    del _get_bbox
    del _set_bbox

    def _get_size(self):
        """
        Size of the request in pixel as a tuple (width, height),
        or None if one is missing.
        """
        if 'imageDisplay' not in self.params or self.params['imageDisplay'] is None:
            return None
        dim = [float(val) for val in self.params['imageDisplay'].split(',')]
        return tuple(dim[:2])

    def _set_size(self, value):
        if value is not None and not isinstance(value, str):
            value = ','.join(str(x) for x in value) + ',96'
        self['imageDisplay'] = value
    size = property(_get_size, _set_size)
    del _get_size
    del _set_size

    def _get_pos(self):
        size = self.size
        vals = self['geometry'].split(',')
        x, y = float(vals[0]), float(vals[1])
        return make_lin_transf(self.bbox, (0, 0, size[0], size[1]))((x, y))

    def _set_pos(self, value):
        size = self.size
        req_coord = make_lin_transf((0, 0, size[0], size[1]), self.bbox)(value)
        self['geometry'] = '%f,%f' % req_coord
    pos = property(_get_pos, _set_pos)
    del _get_pos
    del _set_pos

    @property
    def srs(self):
        srs = self.params.get('sr', None)
        if srs:
            return 'EPSG:%s' % srs

    @srs.setter
    def srs(self, srs):
        if hasattr(srs, 'srs_code'):
            code = srs.srs_code
        else:
            code = srs
        self.params['sr'] = code.rsplit(':', 1)[-1]


class ArcGISRequest(BaseRequest):
    request_params = ArcGISExportRequestParams
    fixed_params = {"f": "image"}

    def __init__(self, param=None, url='', validate=False, http=None):
        BaseRequest.__init__(self, param, url, validate, http)

        self.url = rest_endpoint(url)

    def copy(self):
        return self.__class__(param=self.params.copy(), url=self.url)

    @property
    def query_string(self):
        params = self.params.copy()
        for key, value in self.fixed_params.items():
            params[key] = value
        return params.query_string


class ArcGISIdentifyRequest(BaseRequest):
    request_params = ArcGISIdentifyRequestParams
    fixed_params = {'geometryType': 'esriGeometryPoint'}

    def __init__(self, param=None, url='', validate=False, http=None):
        BaseRequest.__init__(self, param, url, validate, http)

        self.url = rest_identify_endpoint(url)

    def copy(self):
        return self.__class__(param=self.params.copy(), url=self.url)

    @property
    def query_string(self):
        params = self.params.copy()
        for key, value in self.fixed_params.items():
            params[key] = value
        return params.query_string


def create_identify_request(req_data, param):
    req_data = req_data.copy()

    # Pop the URL off the request data.
    url = req_data['url']
    del req_data['url']

    return ArcGISIdentifyRequest(url=url, param=req_data)


def create_request(req_data, param):
    req_data = req_data.copy()

    # Pop the URL off the request data.
    url = req_data['url']
    del req_data['url']

    if 'format' in param:
        req_data['format'] = param['format']

    if 'transparent' in req_data:
        # Convert boolean to a string.
        req_data['transparent'] = str(req_data['transparent'])

    return ArcGISRequest(url=url, param=req_data)


def rest_endpoint(url):
    parts = urlsplit(url)
    path = parts.path.rstrip('/').split('/')

    if path[-1] in ('export', 'exportImage'):
        if path[-2] == 'MapServer':
            path[-1] = 'export'
        elif path[-2] == 'ImageServer':
            path[-1] = 'exportImage'
    elif path[-1] == 'MapServer':
        path.append('export')
    elif path[-1] == 'ImageServer':
        path.append('exportImage')

    parts = parts[0], parts[1], '/'.join(path), parts[3], parts[4]
    return urlunsplit(parts)


def rest_identify_endpoint(url):
    parts = urlsplit(url)
    path = parts.path.rstrip('/').split('/')

    if path[-1] in ('export', 'exportImage'):
        path[-1] = 'identify'
    elif path[-1] in ('MapServer', 'ImageServer'):
        path.append('identify')

    parts = parts[0], parts[1], '/'.join(path), parts[3], parts[4]
    return urlunsplit(parts)



##############################################################################
#
# This is a modified version of zc.lockfile 1.0.0
# (http://pypi.python.org/pypi/zc.lockfile/1.0.0)
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# ==== Changelog ====
# 2010-04-01 - Commented out logging. <olt@omniscale.de>
#
# ==== License ====
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).
#
# Zope Public License (ZPL) Version 2.1
#
# A copyright notice accompanies this license document that identifies the
# copyright holders.
#
# This license has been certified as open source. It has also been designated as
# GPL compatible by the Free Software Foundation (FSF).
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# - Redistributions in source code must retain the accompanying copyright
#   notice, this list of conditions, and the following disclaimer.
#
# - Redistributions in binary form must reproduce the accompanying copyright
#   notice, this list of conditions, and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# - Names of the copyright holders must not be used to endorse or promote
#   products derived from this software without prior written permission from
#   the copyright holders.
#
# - The right to distribute this software or to use it for any purpose does not
#   give you the right to use Servicemarks (sm) or Trademarks (tm) of the
#   copyright holders. Use of them is covered by separate agreement with the
#   copyright holders.
#
# - If any files are modified, you must cause the modified files to carry
#   prominent notices stating that you changed the files and the date of any
#   change.
#
# Disclaimer
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY EXPRESSED
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
# EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
##############################################################################

import os
# import logging
# logger = logging.getLogger("zc.lockfile")


class LockError(Exception):
    """Couldn't get a lock
    """


try:
    import fcntl
except ImportError:
    try:
        import msvcrt
    except ImportError:
        def _lock_file(file):
            raise TypeError('No file-locking support on this platform')

        def _unlock_file(file):
            raise TypeError('No file-locking support on this platform')

    else:
        # Windows
        def _lock_file(file):
            # Lock just the first byte
            try:
                msvcrt.locking(file.fileno(), msvcrt.LK_NBLCK, 1)
            except IOError:
                raise LockError("Couldn't lock %r" % file.name)

        def _unlock_file(file):
            try:
                file.seek(0)
                msvcrt.locking(file.fileno(), msvcrt.LK_UNLCK, 1)
            except IOError:
                raise LockError("Couldn't unlock %r" % file.name)

else:
    # Unix
    _flags = fcntl.LOCK_EX | fcntl.LOCK_NB

    def _lock_file(file):
        try:
            fcntl.flock(file.fileno(), _flags)
        except (IOError, OSError) as err:
            raise LockError("Couldn't lock {0}, error: {1}".format(file.name, err))

    def _unlock_file(file):
        # File is automatically unlocked on close
        pass


class LockFile:

    _fp = None

    def __init__(self, path, file_permissions):
        self._path = path
        set_permissions = file_permissions and not os.path.exists(path)
        try:
            fp = open(path, 'w+')
        except IOError:
            raise Exception('Could not create Lock-file, wrong permissions on lock directory?')

        if set_permissions:
            permission = int(file_permissions, base=8)
            os.chmod(path, permission)

        try:
            _lock_file(fp)
        except Exception as ex:
            try:
                fp.close()
            except Exception:
                pass
            raise ex

        self._fp = fp
        fp.write(" %s\n" % os.getpid())
        fp.truncate()
        fp.flush()

    def close(self):
        if self._fp is not None:
            _unlock_file(self._fp)
            self._fp.close()
            self._fp = None



from __future__ import print_function
import struct
from mapproxy.cache.base import tile_buffer
from mapproxy.image import ImageSource


class MetaTileFile(object):
    def __init__(self, meta_tile):
        self.meta_tile = meta_tile

    def write_tiles(self, tiles):
        tile_positions = []
        count = len(tiles)  # self.meta_tile.grid_size[0]
        header_size = (
            4   # META
            + 4   # metasize**2
            + 3*4  # x, y, z
            + count * 8  # offset/size * tiles
        )
        with open('/tmp/foo.metatile', 'wb') as f:
            f.write("META")
            f.write(struct.pack('i', count))
            f.write(struct.pack('iii', *tiles[0].coord))
            offsets_header_pos = f.tell()
            f.seek(header_size, 0)

            for tile in tiles:
                offset = f.tell()
                with tile_buffer(tile) as buf:
                    tile_data = buf.read()
                    f.write(tile_data)
                tile_positions.append((offset, len(tile_data)))

            f.seek(offsets_header_pos, 0)
            for offset, size in tile_positions:
                f.write(struct.pack('ii', offset, size))

    def _read_header(self, f):
        f.seek(0, 0)
        assert f.read(4) == "META"
        count, x, y, z = struct.unpack('iiii', f.read(4*4))
        tile_positions = []
        for i in range(count):
            offset, size = struct.unpack('ii', f.read(4*2))
            tile_positions.append((offset, size))

        return tile_positions

    def read_tiles(self):
        with open('/tmp/foo.metatile', 'rb') as f:
            tile_positions = self._read_header(f)

            for i, (offset, size) in enumerate(tile_positions):
                f.seek(offset, 0)
                # img = ImageSource(BytesIO(f.read(size)))
                open('/tmp/img-%02d.png' % i, 'wb').write(f.read(size))


if __name__ == '__main__':
    from io import BytesIO
    from mapproxy.cache.tile import Tile
    from mapproxy.test.image import create_tmp_image

    tiles = []
    img = create_tmp_image((256, 256))
    for x in range(8):
        for y in range(8):
            tiles.append(Tile((x, y, 4), ImageSource(BytesIO(img))))

    m = MetaTileFile(None)
    print('!')
    m.write_tiles(tiles)
    print('!')
    m.read_tiles()
    print('!')

    x = y = 0
    METATILE = 8
    for meta in range(METATILE ** 2):
        print(x + (meta / METATILE), y + (meta % METATILE))



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from io import BytesIO

import pytest

from PIL import Image
from mapproxy.test.http import mock_httpd
from mapproxy.test.image import tmp_image
from mapproxy.test.system import SysTest
from mapproxy.source import ogcapitiles


class TestOGCAPITilesSource(SysTest):
    @pytest.fixture(scope="class")
    def config_file(self):
        return "ogcapitiles_source.yaml"

    def test_global_same_tiling_scheme(self, app):
        ogcapitiles.reset_config_cache = True

        landing_page = {
            "links": [
                {
                    "rel": "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                    "type": "application/json",
                    "title": "Map tilesets available for this dataset (as JSON)",
                    "href": "/ogcapi/map/tiles?f=json",
                },
            ]
        }

        tilesets = {
            "tilesets": [
                {
                    "title": "my_collection",
                    "tileMatrixSetURI": "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad",
                    "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                    "dataType": "map",
                    "links": [
                        {
                            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
                            "type": "application/json",
                            "title": "WebMercatorQuadTileMatrixSet definition (as JSON)",
                            "href": "/ogcapi/tileMatrixSets/WebMercatorQuad",
                        },
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "WebMercatorQuad map tileset for my_collection (as JSON)",
                            "href": "/ogcapi/map/tiles/WebMercatorQuad?f=json",
                        },
                    ],
                },
            ]
        }

        WebMercatorQuad = {
            "id": "WebMercatorQuad",
            "title": "WebMercatorQuad",
            "uri": "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad",
            "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
            "orderedAxes": ["E", "N"],
            "wellKnownScaleSet": "http://www.opengis.net/def/wkss/OGC/1.0/GoogleMapsCompatible",
            "tileMatrices": [
                {
                    "id": "0",
                    "scaleDenominator": 559082264.028718,
                    "cellSize": 156543.03392804,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 1,
                    "matrixHeight": 1,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
                {
                    "id": "1",
                    "scaleDenominator": 279541132.01436,
                    "cellSize": 78271.5169640205,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 2,
                    "matrixHeight": 2,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
                {
                    "id": "2",
                    "scaleDenominator": 139770566.00718,
                    "cellSize": 39135.7584820102,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 4,
                    "matrixHeight": 4,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
            ],
        }

        tiles = {
            "title": "dataset",
            "tileMatrixSetURI": "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad",
            "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
            "dataType": "map",
            "links": [
                {
                    "rel": "item",
                    "type": "image/png",
                    "title": "WebMercatorQuad map tiles for dataset (as PNG)",
                    "href": "/ogcapi/map/tiles/WebMercatorQuad/{tileMatrix}/{tileRow}/{tileCol}.png",
                    "templated": True,
                },
            ],
        }

        with tmp_image((256, 256), format="png", color=(255, 0, 0)) as img:
            expected_reqs = [
                (
                    {"path": r"/ogcapi"},
                    {
                        "body": bytes(json.dumps(landing_page).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/map/tiles?f=json"},
                    {
                        "body": bytes(json.dumps(tilesets).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/tileMatrixSets/WebMercatorQuad"},
                    {
                        "body": bytes(json.dumps(WebMercatorQuad).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/map/tiles/WebMercatorQuad?f=json"},
                    {
                        "body": bytes(json.dumps(tiles).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/map/tiles/WebMercatorQuad/2/0/1.png"},
                    {"body": img.read(), "headers": {"content-type": "image/png"}},
                ),
            ]
            with mock_httpd(("localhost", 42423), expected_reqs):
                resp = app.get("/tiles/test_cache_global/webmercator/2/1/0.png")
                assert resp.content_type == "image/png"
                img = Image.open(BytesIO(resp.body))
                assert img.format == "PNG"
                assert img.width == 256
                assert img.height == 256
                assert img.getextrema() == ((255, 255), (0, 0), (0, 0))

        with tmp_image((256, 256), format="png", color=(0, 255, 0)) as img:
            expected_reqs = [
                (
                    {"path": r"/ogcapi/map/tiles/WebMercatorQuad/0/0/0.png"},
                    {"body": img.read(), "headers": {"content-type": "image/png"}},
                ),
            ]
            with mock_httpd(("localhost", 42423), expected_reqs):
                resp = app.get("/tiles/test_cache_global/webmercator/0/0/0.png")
                assert resp.content_type == "image/png"
                img = Image.open(BytesIO(resp.body))
                assert img.format == "PNG"
                assert img.width == 256
                assert img.height == 256
                assert img.getextrema() == ((0, 0), (255, 255), (0, 0))

    def test_collection_not_same_tiling_scheme(self, app):
        ogcapitiles.reset_config_cache = True

        my_collection = {
            "links": [
                {
                    "rel": "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                    "type": "application/json",
                    "title": "Map tilesets available for this dataset (as JSON)",
                    "href": "/ogcapi/collections/my_collection/map/tiles?f=json",
                },
            ]
        }

        tilesets = {
            "tilesets": [
                {
                    "title": "my_collection",
                    "tileMatrixSetURI": "http://www.opengis.net/def/tilematrixset/OGC/1.0/tileMatrixSet",
                    "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                    "dataType": "map",
                    "links": [
                        {
                            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
                            "type": "application/json",
                            "title": "tileMatrixSet definition (as JSON)",
                            "href": "/ogcapi/tileMatrixSets/tileMatrixSet",
                        },
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "tileMatrixSet map tileset for my_collection (as JSON)",
                            "href": "/ogcapi/collections/my_collection/map/tiles/tileMatrixSet?f=json",
                        },
                    ],
                },
                {
                    "title": "my_collection",
                    "tileMatrixSetURI": "http://www.opengis.net/def/tilematrixset/OGC/1.0/my_tms",
                    "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
                    "dataType": "map",
                    "links": [
                        {
                            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
                            "type": "application/json",
                            "title": "my_tms definition (as JSON)",
                            "href": "/ogcapi/tileMatrixSets/my_tms",
                        },
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "my_tms map tileset for my_collection (as JSON)",
                            "href": "/ogcapi/collections/my_collection/map/tiles/my_tms?f=json",
                        },
                    ],
                },
            ]
        }

        my_tms = {
            "id": "my_tms",
            "title": "my_tms",
            "uri": "http://www.opengis.net/def/tilematrixset/OGC/1.0/my_tms",
            "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
            "orderedAxes": ["E", "N"],
            "tileMatrices": [
                {
                    "id": "0",
                    "scaleDenominator": 279541132.01436,
                    "cellSize": 78271.5169640205,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [-20037508.3427892, 20037508.3427892],
                    "matrixWidth": 2,
                    "matrixHeight": 2,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
            ],
        }

        tiles = {
            "title": "my_collection",
            "tileMatrixSetURI": "http://www.opengis.net/def/tilematrixset/OGC/1.0/my_tms",
            "crs": "http://www.opengis.net/def/crs/EPSG/0/3857",
            "dataType": "map",
            "links": [
                {
                    "rel": "item",
                    "type": "image/png",
                    "title": "my_tms map tiles for my_collection (as PNG)",
                    "href": "/ogcapi/collections/my_collection/map/tiles/my_tms/{tileMatrix}/{tileRow}/{tileCol}.png",
                    "templated": True,
                },
            ],
        }

        expected_reqs = [
            (
                {"path": r"/ogcapi/collections/my_collection"},
                {
                    "body": bytes(json.dumps(my_collection).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/collections/my_collection/map/tiles?f=json"},
                {
                    "body": bytes(json.dumps(tilesets).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/tileMatrixSets/my_tms"},
                {
                    "body": bytes(json.dumps(my_tms).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
            (
                {"path": r"/ogcapi/collections/my_collection/map/tiles/my_tms?f=json"},
                {
                    "body": bytes(json.dumps(tiles).encode("UTF-8")),
                    "headers": {"content-type": "application/json"},
                },
            ),
        ]
        with mock_httpd(("localhost", 42423), expected_reqs, unordered=True):
            resp = app.get(
                "/tiles/test_cache/webmercator/0/0/0.png", expect_errors=True
            )
            assert resp.content_type == "text/plain"
            assert resp.body == b"BBOX does not align to tile"


class TestOGCAPITilesNonEarthSource(SysTest):
    @pytest.fixture(scope="class")
    def config_file(self):
        return "ogcapitiles_non_earth_source.yaml"

    def test(self, app):
        ogcapitiles.reset_cache = True

        my_collection = {
            "links": [
                {
                    "rel": "http://www.opengis.net/def/rel/ogc/1.0/tilesets-map",
                    "type": "application/json",
                    "title": "Map tilesets available for this dataset (as JSON)",
                    "href": "/ogcapi/collections/my_collection/map/tiles?f=json",
                },
            ],
            "storageCrs": "http://www.opengis.net/def/crs/IAU/2015/30100",
        }

        tilesets = {
            "tilesets": [
                {
                    "title": "my_collection",
                    "crs": "http://www.opengis.net/def/crs/IAU/2015/30100",
                    "dataType": "map",
                    "links": [
                        {
                            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
                            "type": "application/json",
                            "title": "moon_grid definition (as JSON)",
                            "href": "/ogcapi/tileMatrixSets/moon_grid",
                        },
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "moon_grid map tileset for my_collection (as JSON)",
                            "href": "/ogcapi/collections/my_collection/map/tiles/moon_grid?f=json",
                        },
                    ],
                },
                {
                    "title": "my_collection",
                    "crs": "http://www.opengis.net/def/crs/IAU/2015/30100",
                    "dataType": "map",
                    "links": [
                        {
                            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
                            "type": "application/json",
                            "title": "moon_grid definition (as JSON)",
                            "href": "/ogcapi/tileMatrixSets/moon_grid",
                        },
                        {
                            "rel": "self",
                            "type": "application/json",
                            "title": "moon_grid map tileset for my_collection (as JSON)",
                            "href": "/ogcapi/collections/my_collection/map/tiles/moon_grid?f=json",
                        },
                    ],
                },
            ]
        }

        moon_grid = {
            "id": "moon_grid",
            "title": "moon_grid",
            "crs": "http://www.opengis.net/def/crs/IAU/2015/30100",
            "orderedAxes": ["N", "E"],
            "tileMatrices": [
                {
                    "id": "0",
                    "scaleDenominator": 559082264.0287179,
                    "cellSize": 1.40625,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [90, -180],
                    "matrixWidth": 1,
                    "matrixHeight": 1,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
                {
                    "id": "1",
                    "scaleDenominator": 279541132.01435894,
                    "cellSize": 0.703125,
                    "cornerOfOrigin": "topLeft",
                    "pointOfOrigin": [90, -180],
                    "matrixWidth": 2,
                    "matrixHeight": 2,
                    "tileWidth": 256,
                    "tileHeight": 256,
                },
            ],
        }

        tiles = {
            "title": "my_collection",
            "crs": "http://www.opengis.net/def/crs/IAU/2015/30100",
            "dataType": "map",
            "links": [
                {
                    "rel": "item",
                    "type": "image/png",
                    "title": "moon_grid map tiles for my_collection (as PNG)",
                    "href": "/ogcapi/collections/my_collection/map/tiles/moon_grid/{tileMatrix}/{tileRow}/{tileCol}.png",  # noqa
                    "templated": True,
                },
            ],
        }

        with tmp_image((512, 256), format="png", color=(255, 0, 0)) as img:
            expected_reqs = [
                (
                    {"path": r"/ogcapi/collections/my_collection"},
                    {
                        "body": bytes(json.dumps(my_collection).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/collections/my_collection/map/tiles?f=json"},
                    {
                        "body": bytes(json.dumps(tilesets).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {"path": r"/ogcapi/tileMatrixSets/moon_grid"},
                    {
                        "body": bytes(json.dumps(moon_grid).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {
                        "path": r"/ogcapi/collections/my_collection/map/tiles/moon_grid?f=json"
                    },
                    {
                        "body": bytes(json.dumps(tiles).encode("UTF-8")),
                        "headers": {"content-type": "application/json"},
                    },
                ),
                (
                    {
                        "path": r"/ogcapi/collections/my_collection/map/tiles/moon_grid/1/0/1.png"
                    },
                    {"body": img.read(), "headers": {"content-type": "image/png"}},
                ),
            ]
            with mock_httpd(("localhost", 42423), expected_reqs):
                resp = app.get("/wmts/test/moon_grid/1/1/0.png")
                assert resp.content_type == "image/png"
                img = Image.open(BytesIO(resp.body))
                assert img.format == "PNG"
                assert img.width == 512
                assert img.height == 256
                assert img.getextrema() == ((255, 255), (0, 0), (0, 0))



# This file is part of the MapProxy project.
# Copyright (C) 2015 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import json
import os.path
from typing import List, Dict, Set, Tuple, Union, Iterable

from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft202012Validator

import mapproxy.config.defaults

import logging
log = logging.getLogger('mapproxy.config')


with open(os.path.join(os.path.dirname(__file__), 'config-schema.json')) as schema_file:
    schema = json.load(schema_file)


TAGGED_SOURCE_TYPES = [
    'wms',
    'mapserver',
    'mapnik'
]


def get_error_messages(errors: Iterable[ValidationError]) -> List[str]:
    msgs = []
    for error in errors:
        path = error.json_path.replace('$', 'root')
        msg = f'{error.message} in {path}'
        msgs.append(msg)
        if error.context is not None:
            msgs += get_error_messages(error.context)
    return msgs


def validate(conf_dict: Dict) -> List[str]:
    validator = Draft202012Validator(schema=schema)
    errors_iter = validator.iter_errors(conf_dict)
    errors = [] if errors_iter is None else get_error_messages(errors_iter)

    layers_conf = conf_dict.get('layers')

    if layers_conf is None:
        return errors
    else:
        return errors + list(itertools.chain.from_iterable(_validate_layer(conf_dict, layer) for layer in layers_conf))


def _validate_layer(conf_dict: Dict, layer: Dict) -> List[str]:
    layer_sources: List[str] = layer.get('sources', [])
    tile_sources: List[str] = layer.get('tile_sources', [])
    child_layers: List[Dict] = layer.get('layers', [])

    caches_conf: Dict = conf_dict.get('caches', {})
    sources_conf: Dict = conf_dict.get('sources', {})

    errors = []

    for child_layer in child_layers:
        errors += _validate_layer(conf_dict, child_layer)

    for source in layer_sources:
        if source in caches_conf:
            errors += _validate_cache(conf_dict, source, caches_conf[source])
            continue

        source, layers = _split_tagged_source(source)
        if source in sources_conf:
            errors += _validate_source(conf_dict, source, sources_conf[source], layers)
            continue

        errors.append(
            f"Source '{source}' for layer '{layer['name']}' not in cache or source section"
        )

    for source in tile_sources:
        if source in caches_conf:
            errors += _validate_cache(conf_dict, source, caches_conf[source])
            continue

        errors.append(
            f"Tile source '{source}' for layer '{layer['name']}' not in cache section"
        )

    return errors


def _split_tagged_source(source_name: str) -> Tuple[str, List[str]]:
    layers = None
    if ':' in str(source_name):
        source_name, layers = str(source_name).split(':', 1)
        layers = layers.split(',') if layers is not None else None
    return source_name, layers


def _validate_source(conf_dict: Dict, name: str, source: Dict, layers: List[str]) -> List[str]:
    source_type = source.get('type')
    if source_type == 'wms':
        return _validate_wms_source(name, source, layers)
    elif source_type == 'mapserver':
        return _validate_mapserver_source(conf_dict, name, source, layers)
    elif source_type == 'mapnik':
        return _validate_mapnik_source(name, source, layers)
    return []


def _validate_wms_source(name: str, source: Dict, layers: List[str]) -> List[str]:
    errors = []
    if source['req'].get('layers') is None and layers is None:
        errors.append("Missing 'layers' for source '%s'" % (
            name
        ))
    if source['req'].get('layers') is not None and layers is not None:
        errors += _validate_tagged_layer_source(
            name,
            source['req'].get('layers'),
            layers
        )
    return errors


def _validate_mapserver_source(conf_dict: Dict, name: str, source: Dict, layers: List[str]) -> List[str]:
    globals_conf: Dict = conf_dict.get('globals')
    errors = []
    mapserver: Dict = source.get('mapserver')
    if mapserver is None:
        if (
                not globals_conf or
                not globals_conf.get('mapserver') or
                not globals_conf['mapserver'].get('binary')
        ):
            errors.append(f"Missing mapserver binary for source '{name}'")
        elif not os.path.isfile(globals_conf['mapserver']['binary']):
            errors.append(f"Could not find mapserver binary ({globals_conf['mapserver'].get('binary')})")
    elif mapserver is None or not source['mapserver'].get('binary'):
        errors.append(f"Missing mapserver binary for source '{name}'")
    elif not os.path.isfile(source['mapserver']['binary']):
        errors.append(f"Could not find mapserver binary ({source['mapserver']['binary']})")

    if source['req'].get('layers') and layers is not None:
        errors += _validate_tagged_layer_source(
            name,
            source['req'].get('layers'),
            layers
        )

    return errors


def _validate_mapnik_source(name: str, source: Dict, layers: List[str]) -> List[str]:
    if source.get('layers') and layers is not None:
        return _validate_tagged_layer_source(name, source.get('layers'), layers)
    return []


def _validate_tagged_layer_source(
        name: str, supported_layers: Union[str, List[str]], requested_layers: List[str]) -> List[str]:
    errors = []
    if isinstance(supported_layers, str):
        supported_layers = supported_layers.split(',')
    if not set(requested_layers).issubset(set(supported_layers)):
        return [
            f"Supported layers for source '{name}' are '{', '.join(supported_layers)}' but tagged source requested"
            f" layers '{', '.join(requested_layers)}'"
        ]
    return errors


def _validate_cache(conf_dict: Dict, name: str, cache: Dict) -> List[str]:
    errors = []
    if isinstance(cache.get('sources', []), dict):
        errors += _validate_bands(name, set(cache['sources'].keys()))
        for band, confs in cache['sources'].items():
            for conf in confs:
                band_source = conf['source']
                errors += _validate_cache_source(conf_dict, name, band_source)
    else:
        for cache_source in cache.get('sources', []):
            errors += _validate_cache_source(conf_dict, name, cache_source)

    for grid in cache.get('grids', []):
        if grid not in _get_known_grids(conf_dict):
            errors.append(
                f"Grid '{grid}' for cache '{name}' not found in config"
            )

    return errors


def _validate_cache_source(conf_dict: Dict, cache_name: str, source_name: str) -> List[str]:
    errors = []
    sources_conf: Dict = conf_dict.get('sources', {})
    caches_conf: Dict = conf_dict.get('caches', {})
    source_name, layers = _split_tagged_source(source_name)
    if sources_conf and source_name in sources_conf:
        source = sources_conf.get(source_name)
        if layers is not None and source.get('type') not in TAGGED_SOURCE_TYPES:
            errors.append(
                f"Found tagged source '{source_name}' in cache '{cache_name}' but tagged sources only supported for"
                f" '{', '.join(TAGGED_SOURCE_TYPES)}' sources"
            )
            return errors
        errors += _validate_source(conf_dict, source_name, source, layers)
        return errors
    if caches_conf and source_name in caches_conf:
        errors += _validate_cache(conf_dict, source_name, caches_conf[source_name])
        return errors
    errors.append(
        f"Source '{source_name}' for cache '{cache_name}' not found in config"
    )
    return errors


def _validate_bands(cache_name: str, bands: Set[str]) -> List[str]:
    if 'l' in bands and len(bands) > 1:
        return [
            f"Cannot combine 'l' band with bands in cache '{cache_name}'"
        ]
    return []


def _get_known_grids(conf_dict: Dict) -> Set[str]:
    grids_conf = conf_dict.get('grids')
    known_grids = set(mapproxy.config.defaults.grids.keys())
    if grids_conf:
        known_grids.update(grids_conf.keys())
    return known_grids



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

import pytest

from mapproxy.test.system import SysTest
from mapproxy.test.system.test_wms import (
    is_111_capa,
    is_110_capa,
    is_100_capa,
    is_130_capa,
    ns130,
)


TESTSERVER_ADDRESS = "localhost", 42423


class TestWMSWithRoot(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "layergroups_root.yaml"

    def _check_layernames(self, xml):
        assert xml.xpath("//Capability/Layer/Title/text()") == ["Root Layer"]
        assert xml.xpath("//Capability/Layer/Name/text()") == ["root"]
        assert xml.xpath("//Capability/Layer/Layer/Name/text()") == ["layer1", "layer2"]
        assert xml.xpath("//Capability/Layer/Layer[1]/Layer/Name/text()") == [
            "layer1a",
            "layer1b",
        ]

    def _check_layernames_with_namespace(self, xml, namespaces=None):
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Title/text()", namespaces=namespaces
        ) == ["Root Layer"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Name/text()", namespaces=namespaces
        ) == ["root"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer/wms:Name/text()",
            namespaces=namespaces,
        ) == ["layer1", "layer2"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer[1]/wms:Layer/wms:Name/text()",
            namespaces=namespaces,
        ) == ["layer1a", "layer1b"]

    def test_100_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&wmtver=1.0.0")
        xml = resp.lxml
        assert is_100_capa(xml)
        self._check_layernames(xml)

    def test_110_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.1.0")
        xml = resp.lxml
        assert is_110_capa(xml)
        self._check_layernames(xml)

    def test_111_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.1.1")
        xml = resp.lxml
        assert is_111_capa(xml)
        self._check_layernames(xml)

    def test_130_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.3.0")
        xml = resp.lxml
        assert is_130_capa(xml)
        self._check_layernames_with_namespace(xml, ns130)


class TestWMSWithoutRoot(SysTest):

    @pytest.fixture(scope="class")
    def config_file(self):
        return "layergroups.yaml"

    def _check_layernames(self, xml):
        assert xml.xpath("//Capability/Layer/Title/text()") == ["My WMS"]
        assert xml.xpath("//Capability/Layer/Name/text()") == []
        assert xml.xpath("//Capability/Layer/Layer/Name/text()") == ["layer1", "layer2"]
        assert xml.xpath("//Capability/Layer/Layer[1]/Layer/Name/text()") == [
            "layer1a",
            "layer1b",
        ]
        assert xml.xpath("//Capability/Layer/Layer[2]/Layer/Name/text()") == [
            "layer2a",
            "layer2b",
        ]
        assert xml.xpath("//Capability/Layer/Layer[2]/Layer/Layer[1]/Name/text()") == [
            "layer2b1"
        ]

    def _check_layernames_with_namespace(self, xml, namespaces=None):
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Title/text()", namespaces=namespaces
        ) == ["My WMS"]
        assert (
            xml.xpath(
                "//wms:Capability/wms:Layer/wms:Name/text()", namespaces=namespaces
            )
            == []
        )
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer/wms:Name/text()",
            namespaces=namespaces,
        ), ["layer1" == "layer2"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer[1]/wms:Layer/wms:Name/text()",
            namespaces=namespaces,
        ) == ["layer1a", "layer1b"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer[2]/wms:Layer/wms:Name/text()",
            namespaces=namespaces,
        ) == ["layer2a", "layer2b"]
        assert xml.xpath(
            "//wms:Capability/wms:Layer/wms:Layer[2]/wms:Layer/wms:Layer[1]/wms:Name/text()",
            namespaces=namespaces,
        ) == ["layer2b1"]

    def test_100_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&wmtver=1.0.0")
        xml = resp.lxml
        assert is_100_capa(xml)
        self._check_layernames(xml)

    def test_110_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.1.0")
        xml = resp.lxml
        assert is_110_capa(xml)
        self._check_layernames(xml)

    def test_111_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.1.1")
        xml = resp.lxml
        assert is_111_capa(xml)
        self._check_layernames(xml)

    def test_130_capa(self, app):
        resp = app.get("/service?request=GetCapabilities&service=WMS&version=1.3.0")
        xml = resp.lxml
        assert is_130_capa(xml)
        self._check_layernames_with_namespace(xml, ns130)



# This file is part of the MapProxy project.
# Copyright (C) 2010, 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mapproxy.util.times import timestamp_from_isodate

import pytest


def test_timestamp_from_isodate():
    ts = timestamp_from_isodate("2009-06-09T10:57:00")
    assert (1244537820.0 - 14 * 3600) < ts < (1244537820.0 + 14 * 3600)

    with pytest.raises(ValueError):
        timestamp_from_isodate("2009-06-09T10:57")



# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.exception import RequestError
from mapproxy.request.wmts import wmts_request, WMTS100CapabilitiesRequest
from mapproxy.request.wmts import (
    URLTemplateConverter,
    FeatureInfoURLTemplateConverter,
    InvalidWMTSTemplate,
    make_wmts_rest_request_parser,
    WMTS100RestTileRequest,
    WMTS100RestFeatureInfoRequest,
)
from mapproxy.request.base import url_decode

import pytest


def dummy_req(args):
    return DummyRequest(url_decode(args.replace("\n", "")))


def dummy_rest_req(path):
    return DummyRequest(args="", path=path)


class DummyRequest(object):

    def __init__(self, args, path="", url=""):
        self.args = args
        self.path = path
        self.base_url = url


def test_tile_request():
    args = """request=GetTile&service=wmts&tileMatrixset=EPSG900913&
tilematrix=2&tileROW=4&TILECOL=2&FORMAT=image/png&Style=&layer=Foo&version=1.0.0"""
    req = wmts_request(dummy_req(args))

    assert req.params.coord == (2, 4, "2")
    assert req.params.layer == "Foo"
    assert req.params.format == "png"
    assert req.params.tilematrixset == "EPSG900913"


def test_featureinfo_request():
    args = """request=GetFeatureInfo&service=wmts&tileMatrixset=EPSG900913&
tilematrix=2&tileROW=4&TILECOL=2&FORMAT=image/png&Style=&layer=Foo&version=1.0.0&
i=5&j=10&infoformat=application/json"""
    req = wmts_request(dummy_req(args))

    assert req.params.coord == (2, 4, "2")
    assert req.params.layer == "Foo"
    assert req.params.format == "png"
    assert req.params.tilematrixset == "EPSG900913"
    assert req.params.pos == (5, 10)
    assert req.params.infoformat == "application/json"


def test_capabilities_request():
    args = """request=GetCapabilities&service=wmts"""
    req = wmts_request(dummy_req(args))

    assert isinstance(req, WMTS100CapabilitiesRequest)


def test_template_converter():
    regexp = URLTemplateConverter(
        "/{Layer}/{Style}/{TileMatrixSet}-{TileMatrix}-{TileCol}-{TileRow}/tile"
    ).regexp()
    match = regexp.match("/wmts/test/bar/foo-EPSG4326-4-12-99/tile")
    assert match
    assert match.groupdict()["Layer"] == "test"
    assert match.groupdict()["TileMatrixSet"] == "foo-EPSG4326"
    assert match.groupdict()["TileMatrix"] == "4"
    assert match.groupdict()["TileCol"] == "12"
    assert match.groupdict()["TileRow"] == "99"
    assert match.groupdict()["Style"] == "bar"


def test_template_converter_deprecated_format():
    # old format that doesn't match the WMTS spec, now deprecated
    regexp = URLTemplateConverter(
        "/{{Layer}}/{{Style}}/{{TileMatrixSet}}-{{TileMatrix}}-{{TileCol}}-{{TileRow}}/tile"
    ).regexp()
    match = regexp.match("/wmts/test/bar/foo-EPSG4326-4-12-99/tile")
    assert match
    assert match.groupdict()["Layer"] == "test"
    assert match.groupdict()["TileMatrixSet"] == "foo-EPSG4326"
    assert match.groupdict()["TileMatrix"] == "4"
    assert match.groupdict()["TileCol"] == "12"
    assert match.groupdict()["TileRow"] == "99"
    assert match.groupdict()["Style"] == "bar"


def test_template_converter_missing_vars():
    with pytest.raises(InvalidWMTSTemplate):
        URLTemplateConverter("/wmts/{Style}/{TileMatrixSet}/{TileCol}.png").regexp()


def test_template_converter_dimensions():
    converter = URLTemplateConverter(
        "/{Layer}/{Dim1}/{Dim2}/{TileMatrixSet}-{TileMatrix}-{TileCol}-{TileRow}/tile"
    )
    assert converter.dimensions == ["Dim1", "Dim2"]


class TestRestRequestParser(object):

    @pytest.fixture
    def parser(self):
        return make_wmts_rest_request_parser(
            URLTemplateConverter(
                "/{Layer}/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}.{Format}"
            ),
            FeatureInfoURLTemplateConverter(
                "/{Layer}/{TileMatrixSet}/{TileMatrix}/{TileCol}/{TileRow}/{I}/{J}.{InfoFormat}"
            ),
        )

    @pytest.mark.parametrize(
        "url,tile,pos",
        [
            ["/wmts/roads/webmercator/09/201/123/10/30.json", (201, 123, 9), (10, 30)],
            ["/wmts/roads/webmercator/10/201/1/0/999.json", (201, 1, 10), (0, 999)],
            ["/wmts/roads/webmercator/09/201/123/10/30json", None, None],
            ["/wmts/roads/webmercator/09/201/123/10.json", None, None],
            ["/wmts/roads-webmercator/09/201/123/10/30.json", None, None],
            ["/roads/webmercator/09/201/123/10/30.json", None, None],
        ],
    )
    def test_featureinfo(self, parser, url, tile, pos):
        if tile is None:
            with pytest.raises(RequestError):
                parser(dummy_rest_req(url))
        else:
            req = parser(dummy_rest_req(url))
            assert isinstance(req, WMTS100RestFeatureInfoRequest)
            req.make_request()
            assert req.pos == pos
            assert req.tile == tile
            assert req.infoformat == "json"
            assert req.tilematrixset == "webmercator"

    @pytest.mark.parametrize(
        "url,tile",
        [
            ["/wmts/roads/webmercator/09/201/123.png", (201, 123, 9)],
            ["/wmts/roads/webmercator/10/201/123.png", (201, 123, 10)],
            ["/wmts/roads/webmercator/10/201/123a.png", None],
            ["/wmts/roads/webmercator/10/201/123png", None],
            ["/wmts/roads/webmercator/10/2013.png", None],
            ["/wmts/roads-webmercator/10/201/123.png", None],
        ],
    )
    def test_tile(self, parser, url, tile):
        if tile is None:
            with pytest.raises(RequestError):
                parser(dummy_rest_req(url))
        else:
            req = parser(dummy_rest_req(url))
            assert isinstance(req, WMTS100RestTileRequest)
            req.make_request()
            assert req.tile == tile
            assert req.format == "png"
            assert req.tilematrixset == "webmercator"



# This file is part of the MapProxy project.
# Copyright (C) 2010, 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Filter for tiles (watermark, etc.)
"""

from mapproxy.image.message import WatermarkImage


def create_watermark_filter(conf, context, **kw):
    text = conf['watermark'].get('text', '')
    opacity = conf['watermark'].get('opacity')
    font_size = conf['watermark'].get('font_size')
    spacing = conf['watermark'].get('spacing')
    font_color = conf['watermark'].get('color')
    if spacing not in ('wide', None):
        raise ValueError('unsupported watermark spacing: %r' % spacing)
    if text != '':
        return watermark_filter(text, opacity=opacity, font_size=font_size,
                                spacing=spacing, font_color=font_color)


def watermark_filter(text, opacity=None, spacing=None, font_size=None, font_color=None):
    """
    Returns a tile filter that adds a watermark to the tiles.
    :param text: watermark text
    """
    def _watermark_filter(tile):
        placement = tile_watermark_placement(tile.coord, spacing == 'wide')
        wimg = WatermarkImage(text, image_opts=tile.source.image_opts,
                              placement=placement, opacity=opacity, font_size=font_size,
                              font_color=font_color)
        tile.source = wimg.draw(img=tile.source, in_place=False)
        return tile
    return _watermark_filter


def tile_watermark_placement(coord, double_spacing=False):
    if not double_spacing:
        if coord[1] % 2 == 0:
            return 'c'
        else:
            return 'b'

    if coord[1] % 2 != coord[0] % 2:
        return 'c'

    return None



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from mapproxy.test.image import tmp_image
from mapproxy.test.http import mock_httpd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "tilesource_minmax_res.yaml"


class TestTileSourceMinMaxRes(SysTest):

    def test_get_tile_res_a(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {"path": r"/tiles_a/06/000/000/000/000/000/001.png"},
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(("localhost", 42423), [expected_req]):
                resp = app.get("/tiles/tms_cache/6/0/1.png")
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "tms_cache_EPSG900913/06/000/000/000/000/000/001.png"
        ).check()

    def test_get_tile_res_b(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {"path": r"/tiles_b/07/000/000/000/000/000/001.png"},
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(("localhost", 42423), [expected_req]):
                resp = app.get("/tiles/tms_cache/7/0/1.png")
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "tms_cache_EPSG900913/07/000/000/000/000/000/001.png"
        ).check()



# -:- encoding: utf-8 -:-
# This file is part of the MapProxy project.
# Copyright (C) 2013 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mapproxy.srs import SRS

import logging


def sources(cap):
    sources = {}
    for layer in cap.layers_list():
        name, conf = for_layer(cap, layer)
        sources[name+'_wms'] = conf

    return sources


_checked_srs = {}


def check_srs(srs):
    if srs not in _checked_srs:
        try:
            SRS(srs)
            _checked_srs[srs] = True
        except Exception as ex:
            logging.getLogger(__name__).warning('unable to initialize srs for %s: %s', srs, ex)
            _checked_srs[srs] = False

    return _checked_srs[srs]


def for_layer(cap, layer):
    source = {'type': 'wms'}

    req = {
        'url': layer['url'],
        'layers': layer['name'],
    }

    if not layer['opaque']:
        req['transparent'] = True

    wms_opts = {}
    if cap.version != '1.1.1':
        wms_opts['version'] = cap.version
    if layer['queryable']:
        wms_opts['featureinfo'] = True
    if layer['legend']:
        wms_opts['legendurl'] = layer['legend']['url']
    if wms_opts:
        source['wms_opts'] = wms_opts

    source['req'] = req

    source['supported_srs'] = []
    for srs in layer['srs']:
        if check_srs(srs):
            source['supported_srs'].append(srs)
    source['supported_srs'].sort()

    if layer['llbbox']:
        source['coverage'] = {
            'srs': 'EPSG:4326',
            'bbox': layer['llbbox'],
        }

    res_hint = layer['res_hint']
    if res_hint:
        if res_hint[0]:
            source['min_res'] = res_hint[0]
        if res_hint[1]:
            source['max_res'] = res_hint[1]

    return layer['name'], source



# This file is part of the MapProxy project.
# Copyright (C) 2025 Spatialys
#
# Initial development funded by Centre National d'Etudes Spatiales (CNES): https://cnes.fr
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from mapproxy.config.config import base_config
from mapproxy.request.base import Request
from mapproxy.srs import SRS
from mapproxy.service.ogcapi.server import OGCAPIException, OGCAPIServer
from mapproxy.service.ogcapi.constants import (
    FORMAT_TYPES,
    F_JSON,
    F_HTML,
    F_PNG,
    F_JPEG,
)
from mapproxy.service.wms import WMSGroupLayer


def _base_tileset(
    server: OGCAPIServer, req: Request, is_html: bool, coll_id: str, tms_name: str
):
    tileset = {}
    tileset["title"] = (
        f"{coll_id} with {tms_name} tile matrix set"
        if coll_id
        else f"Whole dataset with {tms_name} tile matrix set"
    )
    if tms_name == "WebMercatorQuad":
        tileset[
            "tileMatrixSetURI"
        ] = "http://www.opengis.net/def/tilematrixset/OGC/1.0/WebMercatorQuad"
    tileset["dataType"] = "map"

    srs = server.grid_configs[tms_name].tile_grid().srs
    if srs.srs_code == "EPSG:4326":
        srs = SRS("OGC:CRS84")
    elif srs.srs_code == "EPSG:900913":
        srs = SRS("EPSG:3857")
    tileset["crs"] = srs.to_ogc_url()

    if coll_id is None:
        base_url = "/ogcapi"
    else:
        base_url = f"/ogcapi/collections/{coll_id}"

    title_end = f"for {coll_id}" if coll_id else "for the whole dataset"
    tileset["links"] = [
        {
            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
            "type": FORMAT_TYPES[F_JSON],
            "title": f"{tms_name} definition (as JSON)",
            "href": server.create_href(
                req, f"/ogcapi/tileMatrixSets/{tms_name}?f={F_JSON}"
            ),
        },
        {
            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
            "type": FORMAT_TYPES[F_HTML],
            "title": f"{tms_name} definition (as HTML)",
            "href": server.create_href(
                req, f"/ogcapi/tileMatrixSets/{tms_name}?f={F_HTML}"
            ),
        },
        {
            "rel": ("self" if not is_html else "alternate"),
            "type": FORMAT_TYPES[F_JSON],
            "title": f"{tms_name} map tileset {title_end} (as JSON)",
            "href": server.create_href(
                req, f"{base_url}/map/tiles/{tms_name}?f={F_JSON}"
            ),
        },
        {
            "rel": ("alternate" if not is_html else "self"),
            "type": FORMAT_TYPES[F_HTML],
            "title": f"{tms_name} map tileset {title_end} (as HTML)",
            "href": server.create_href(
                req, f"{base_url}/map/tiles/{tms_name}?f={F_HTML}"
            ),
        },
        # Below links are not required in the tileSetsList response (only in
        # full tileSet response), but are provided for convenience
        {
            "rel": "item",
            "type": FORMAT_TYPES[F_PNG],
            "title": f"{tms_name} map tiles {title_end} (as PNG)",
            "href": server.create_href(
                req,
                f"{base_url}/map/tiles/{tms_name}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}.png",
            ),
            "templated": True,
        },
        {
            "rel": "item",
            "type": FORMAT_TYPES[F_JPEG],
            "title": f"{tms_name} map tiles {title_end} (as JPEG)",
            "href": server.create_href(
                req,
                f"{base_url}/map/tiles/{tms_name}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}.jpg",
            ),
            "templated": True,
        },
    ]
    return tileset


def tilesets(server: OGCAPIServer, req: Request, coll_id: Optional[str]):
    log = server.log
    log.debug("TileSets")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    if coll_id is None:
        if server.default_dataset_layers is None:
            raise OGCAPIServer.exception(
                "InternalError",
                "Dataset tilesets request but no default_dataset_layers defined in server configuration",
                status=500,
            )
        base_url = "/ogcapi"
    else:
        if coll_id not in server.layers:
            raise OGCAPIServer.collection_not_found()
        base_url = f"/ogcapi/collections/{coll_id}"

    is_html = server.is_html_req(req)

    title_end = f"for {coll_id}" if coll_id else "for the whole dataset"
    json_resp = {
        "links": [
            {
                "rel": ("self" if not is_html else "alternate"),
                "type": FORMAT_TYPES[F_JSON],
                "title": f"The JSON representation of the available map tilesets {title_end} (as JSON)",
                "href": server.create_href(req, f"{base_url}/map/tiles?f={F_JSON}"),
            },
            {
                "rel": ("alternate" if not is_html else "self"),
                "type": FORMAT_TYPES[F_HTML],
                "title": f"The HTML representation of the available map tilesets {title_end} (as HTML)",
                "href": server.create_href(req, f"{base_url}/map/tiles?f={F_HTML}"),
            },
        ]
    }

    if coll_id:
        layer = server.layers[coll_id]
    else:
        layer = WMSGroupLayer(
            "group_layer", "group_layer", None, server.default_dataset_layers
        )

    tilesets = []
    for tms_name in server.grid_configs:
        is_earth_layer = layer.extent.srs.srs_code.startswith("EPSG:")
        tile_grid_srs = server.grid_configs[tms_name].tile_grid().srs
        is_earth_tilegrid = tile_grid_srs.srs_code.startswith("EPSG:")
        if is_earth_layer != is_earth_tilegrid:
            continue

        tileset = _base_tileset(server, req, is_html, coll_id, tms_name)
        if is_html:
            tileset["tileMatrixSet"] = tms_name

        tilesets.append(tileset)

    if is_html:
        if coll_id:
            json_resp["collections_path"] = server.create_href(
                req, "/ogcapi/collections"
            )
            json_resp["collection_path"] = server.create_href(req, base_url)
        json_resp["collection_title"] = (
            server.layers[coll_id].title if coll_id else "Dataset tiles"
        )
        json_resp["collection_tilesets_path"] = server.create_href(
            req, f"{base_url}/map/tiles"
        )

    json_resp["tilesets"] = tilesets

    return server.create_json_or_html_response(req, json_resp, "tilesets/index.html")


def tileset(server: OGCAPIServer, req: Request, coll_id: str, tms_name: str):
    log = server.log
    log.debug("TileSet")

    for arg in req.args:
        if arg != "f":
            raise OGCAPIServer.unknown_query_parameter(arg)
    if req.args.get("f", None) not in (None, F_JSON, F_HTML):
        raise OGCAPIServer.invalid_parameter("Invalid value for f query parameter")

    if coll_id is None:
        if server.default_dataset_layers is None:
            raise OGCAPIServer.exception(
                "InternalError",
                "Dataset tileset request but no default_dataset_layers defined in server configuration",
                status=500,
            )
        base_url = "/ogcapi"
    else:
        if coll_id not in server.layers:
            raise OGCAPIServer.collection_not_found()
        base_url = f"/ogcapi/collections/{coll_id}"

    if tms_name not in server.grid_configs:
        raise OGCAPIException("Not Found", "TileMatrixSet not found", status=404)

    is_html = server.is_html_req(req)

    json_resp = _base_tileset(server, req, is_html, coll_id, tms_name)

    json_resp["links"] += [
        {
            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
            "type": FORMAT_TYPES[F_JSON],
            "title": f"{tms_name} definition (as JSON)",
            "href": server.create_href(
                req, f"/ogcapi/tileMatrixSets/{tms_name}?f={F_JSON}"
            ),
        },
        {
            "rel": "http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",
            "type": FORMAT_TYPES[F_HTML],
            "title": f"{tms_name} definition (as HTML)",
            "href": server.create_href(
                req, f"/ogcapi/tileMatrixSets/{tms_name}?f={F_HTML}"
            ),
        },
    ]
    if coll_id:
        json_resp["links"] += [
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/geodata",
                "type": FORMAT_TYPES[F_JSON],
                "title": f"{coll_id} collection (as JSON)",
                "href": server.create_href(req, f"{base_url}?f={F_JSON}"),
            },
            {
                "rel": "http://www.opengis.net/def/rel/ogc/1.0/geodata",
                "type": FORMAT_TYPES[F_HTML],
                "title": f"{coll_id} collection (as HTML)",
                "href": server.create_href(req, f"{base_url}?f={F_HTML}"),
            },
        ]

    if is_html:
        if coll_id:
            layer = server.layers[coll_id]
        else:
            layer = WMSGroupLayer(
                "group_layer", "group_layer", None, server.default_dataset_layers
            )

        if coll_id:
            json_resp["collections_path"] = server.create_href(
                req, "/ogcapi/collections"
            )
            json_resp["collection_path"] = server.create_href(req, base_url)
        json_resp["collection_title"] = layer.title
        json_resp["collection_tilesets_path"] = server.create_href(
            req, f"{base_url}/map/tiles"
        )
        json_resp["id"] = tms_name
        json_resp["png_tiles_href"] = server.create_href(
            req,
            f"{base_url}/map/tiles/{tms_name}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}.png",
        )
        json_resp["sample_png_tile_href"] = server.create_href(
            req, f"{base_url}/map/tiles/{tms_name}/0/0/0.png"
        )
        json_resp["jpeg_tiles_href"] = server.create_href(
            req,
            f"{base_url}/map/tiles/{tms_name}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}.jpg",
        )
        json_resp["tms_def_href"] = server.create_href(
            req, f"/ogcapi/tileMatrixSets/{tms_name}"
        )

        tile_grid = server.grid_configs[tms_name].tile_grid()
        srs = tile_grid.srs
        json_resp["srs"] = srs.srs_code
        json_resp["extent"] = list(layer.extent.transform(srs).bbox)
        json_resp["background_url"] = base_config().background.url
        if tile_grid.origin == "ul":
            json_resp["tileset_xyz_url"] = server.create_href(
                req,
                f"{base_url}/map/tiles/{tms_name}/{{z}}/{{y}}/{{x}}.png",
            )
        else:
            json_resp["tileset_xyz_url"] = server.create_href(
                req,
                f"{base_url}/map/tiles/{tms_name}/{{z}}/{{-y}}/{{x}}.png",
            )

    # Recommendation 5.A
    headers = {
        "Link-Template": f'<{base_url}/map/tiles/{tms_name}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}.png>; rel="item"; type="image/png"'  # noqa
    }

    return server.create_json_or_html_response(
        req, json_resp, "tilesets/tileset.html", headers=headers
    )



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The WSGI application.
"""
from __future__ import print_function
import logging
import os
import re
import threading
import time

try:
    # time.strptime is thread-safe, but not the first call.
    # Import _strptime as a workaround. See: http://bugs.python.org/issue7980
    import _strptime  # noqa
except ImportError:
    pass

from mapproxy.request import Request
from mapproxy.response import Response
from mapproxy.config import local_base_config
from mapproxy.config.loader import load_configuration, ConfigurationError
from mapproxy.util.escape import escape_html

log = logging.getLogger('mapproxy.config')
log_wsgiapp = logging.getLogger('mapproxy.wsgiapp')


def make_wsgi_app(services_conf=None, debug=False, ignore_config_warnings=True, reloader=False):
    """
    Create a MapProxyApp with the given services conf.

    :param services_conf: the file name of the mapproxy.yaml configuration
    :param reloader: reload mapproxy.yaml when it changed
    """

    if reloader:
        def make_app():
            return make_wsgi_app(services_conf=services_conf, debug=debug, reloader=False)
        return ReloaderApp(services_conf, make_app)

    try:
        conf = load_configuration(mapproxy_conf=services_conf, ignore_warnings=ignore_config_warnings)
        services = conf.configured_services()
    except ConfigurationError as e:
        log.fatal(e)
        raise

    config_files = conf.config_files()

    app = MapProxyApp(services, conf.base_config)
    if debug:
        app = wrap_wsgi_debug(app, conf)

    app.config_files = config_files
    return app


class ReloaderApp(object):
    def __init__(self, timestamp_file, make_app_func):
        self.timestamp_file = timestamp_file
        self.make_app_func = make_app_func
        self.app = make_app_func()
        self._app_init_lock = threading.Lock()

    def _needs_reload(self):
        for conf_file, timestamp in self.app.config_files.items():
            m_time = os.path.getmtime(conf_file)
            if m_time > timestamp:
                return True
        return False

    def __call__(self, environ, start_response):
        if self._needs_reload():
            with self._app_init_lock:
                if self._needs_reload():
                    try:
                        self.app = self.make_app_func()
                    except ConfigurationError:
                        pass
                    self.last_reload = time.time()

        return self.app(environ, start_response)


def wrap_wsgi_debug(app, conf):
    conf.base_config.debug_mode = True
    try:
        from werkzeug.debug import DebuggedApplication
        app = DebuggedApplication(app, evalex=True)
    except ImportError:
        try:
            from paste.evalexception.middleware import EvalException
            app = EvalException(app)
        except ImportError:
            print('Error: Install Werkzeug or Paste for browser-based debugging.')

    return app


request_interceptors = set()


def register_request_interceptor(interceptor):
    request_interceptors.add(interceptor)


class MapProxyApp(object):
    """
    The MapProxy WSGI application.
    """
    handler_path_re = re.compile(r'^/(\w+)')

    def __init__(self, services, base_config):
        self.handlers = {}
        self.base_config = base_config
        self.cors_origin = base_config.http.access_control_allow_origin
        for service in services:
            for name in service.names:
                self.handlers[name] = service

    def __call__(self, environ, start_response):
        resp = None
        req = Request(environ)

        for interceptor in request_interceptors:
            req = interceptor(req)

        if self.cors_origin:
            orig_start_response = start_response

            def start_response(status, headers, exc_info=None):
                headers.append(('Access-control-allow-origin', self.cors_origin))
                return orig_start_response(status, headers, exc_info)

        with local_base_config(self.base_config):
            match = self.handler_path_re.match(req.path)
            if match:
                handler_name = match.group(1)
                if handler_name in self.handlers:
                    try:
                        resp = self.handlers[handler_name].handle(req)
                    except Exception:
                        if self.base_config.debug_mode:
                            raise
                        else:
                            log_wsgiapp.fatal('fatal error in %s for %s?%s',
                                handler_name, environ.get('PATH_INFO'), environ.get('QUERY_STRING'), exc_info=True)
                            import traceback
                            traceback.print_exc(file=environ['wsgi.errors'])
                            resp = Response('internal error', status=500)
            if resp is None:
                if req.path in ('', '/'):
                    resp = self.welcome_response(escape_html(req.script_url))
                else:
                    resp = Response('not found', mimetype='text/plain', status=404)
            return resp(environ, start_response)

    def welcome_response(self, script_url):
        import mapproxy.version
        html = "<html><body><h1>Welcome to MapProxy %s</h1>" % mapproxy.version.version
        if 'demo' in self.handlers:
            html += '<p>See all configured layers and services at: <a href="%s/demo/">demo</a>' % (script_url, )
        return Response(html, mimetype='text/html')
